# sage.doctest: needs numpy
r"""
Continuous Emission Hidden Markov Models

AUTHOR:

- William Stein, 2010-03
"""

# ***************************************************************************
#       Copyright (C) 2010 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from cpython.object cimport PyObject_RichCompare
from libc.math cimport log, sqrt, exp, isnormal, isfinite, M_PI
cdef double sqrt2pi = sqrt(2*M_PI)
from cysignals.signals cimport sig_on, sig_off

from sage.misc.flatten import flatten
from sage.structure.element import Matrix

from sage.stats.time_series cimport TimeSeries
from sage.stats.intlist cimport IntList

from sage.stats.hmm.hmm cimport HiddenMarkovModel
from sage.stats.hmm.util cimport HMM_Util
from sage.stats.hmm.distributions cimport GaussianMixtureDistribution

cdef HMM_Util util = HMM_Util()

from sage.misc.randstate cimport current_randstate, randstate


# TODO: DELETE THIS FUNCTION WHEN MOVE Gaussian stuff to distributions.pyx!!! (next version)
cdef double random_normal(double mean, double std, randstate rstate) noexcept:
    r"""
    Return a number chosen randomly with given mean and standard deviation.

    INPUT:

    - ``mean`` -- double
    - ``std`` -- double, standard deviation
    - ``rstate`` -- a randstate object

    OUTPUT:

    a double
    """
    # Ported from http://users.tkk.fi/~nbeijar/soft/terrain/source_o2/boxmuller.c
    # This the box muller algorithm.
    # Client code can get the current random state from:
    #         cdef randstate rstate = current_randstate()

    cdef double x1, x2, w, y1, y2
    while True:
        x1 = 2*rstate.c_rand_double() - 1
        x2 = 2*rstate.c_rand_double() - 1
        w = x1*x1 + x2*x2
        if w < 1: break
    w = sqrt( (-2*log(w))/w )
    y1 = x1 * w
    return mean + y1*std

cdef class GaussianHiddenMarkovModel(HiddenMarkovModel):
    r"""
    Gaussian emissions Hidden Markov Model.

    INPUT:

    - ``A`` -- matrix; the `N \times N` transition matrix
    - ``B`` -- list of pairs ``(mu, sigma)`` that define the distributions
    - ``pi`` -- initial state probabilities
    - ``normalize`` -- boolean (default: ``True``)

    EXAMPLES:

    We illustrate the primary functions with an example 2-state Gaussian HMM::

        sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
        ....:                                   [(1,1), (-1,1)],
        ....:                                   [.5,.5]); m
        Gaussian Hidden Markov Model with 2 States
        Transition matrix:
        [0.1 0.9]
        [0.5 0.5]
        Emission parameters:
        [(1.0, 1.0), (-1.0, 1.0)]
        Initial probabilities: [0.5000, 0.5000]

    We query the defining transition matrix, emission parameters, and
    initial state probabilities::

        sage: m.transition_matrix()
        [0.1 0.9]
        [0.5 0.5]
        sage: m.emission_parameters()
        [(1.0, 1.0), (-1.0, 1.0)]
        sage: m.initial_probabilities()
        [0.5000, 0.5000]

    We obtain a sample sequence with 10 entries in it, and compute the
    logarithm of the probability of obtaining this sequence, given the
    model::

        sage: obs = m.sample(5); obs  # random
        [-1.6835, 0.0635, -2.1688, 0.3043, -0.3188]
        sage: log_likelihood = m.log_likelihood(obs)
        sage: counter = 0
        sage: n = 0
        sage: def add_samples(i):
        ....:     global counter, n
        ....:     for _ in range(i):
        ....:         n += 1
        ....:         obs2 = m.sample(5)
        ....:         if all(abs(obs2[i] - obs[i]) < 0.25 for i in range(5)):
        ....:             counter += 1

        sage: add_samples(10000)
        sage: while abs(log_likelihood - log(counter*1.0/n/0.5^5)) < 0.1:
        ....:     add_samples(10000)

    We compute the Viterbi path, and probability that the given path
    of states produced obs::

        sage: m.viterbi(obs)  # random
        ([1, 0, 1, 0, 1], -8.714092684611794)

    We use the Baum-Welch iterative algorithm to find another model
    for which our observation sequence is more likely::

        sage: try:
        ....:     p, s = m.baum_welch(obs)
        ....:     assert p > log_likelihood
        ....:     assert (1 <= s <= 500)
        ....: except RuntimeError:
        ....:     pass

    Notice that running Baum-Welch changed our model::

        sage: m  # random
        Gaussian Hidden Markov Model with 2 States
        Transition matrix:
        [   0.4154981366185841     0.584501863381416]
        [   0.9999993174253741 6.825746258991804e-07]
        Emission parameters:
        [(0.4178882427119503, 0.5173109664360919),
         (-1.5025208631331122, 0.5085512836055119)]
        Initial probabilities: [0.0000, 1.0000]
    """
    cdef TimeSeries B, prob
    cdef int n_out

    def __init__(self, A, B, pi, bint normalize=True):
        r"""
        Create a Gaussian emissions HMM with transition probability
        matrix `A`, normal emissions given by `B`, and initial state
        probability distribution ``pi``.

        INPUT:

        - ``A`` -- list of lists or a square `N \times N` matrix, whose
          `(i,j)` entry gives the probability of transitioning from
          state `i` to state `j`.

        - ``B`` -- list of `N` pairs ``(mu, std)``, where if ``B[i]=(mu,std)``,
          then the probability distribution associated with state `i`
          normal with mean ``mu`` and standard deviation ``std``.

        - ``pi`` -- the probabilities of starting in each initial
          state, i.e., ``pi[i]`` is the probability of starting in
          state `i`.

        - ``normalize`` -- boolean (default: ``True``); if given, input is
          normalized to define valid probability distributions,
          e.g., the entries of `A` are made nonnegative and the rows
          sum to 1.

        EXAMPLES::

            sage: hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]], [(1,1), (-1,1)], [.5,.5])
            Gaussian Hidden Markov Model with 2 States
            Transition matrix:
            [0.1 0.9]
            [0.5 0.5]
            Emission parameters:
            [(1.0, 1.0), (-1.0, 1.0)]
            Initial probabilities: [0.5000, 0.5000]

        We input a model in which both `A` and ``pi`` have to be
        renormalized to define valid probability distributions::

            sage: hmm.GaussianHiddenMarkovModel([[-1,.7],[.3,.4]], [(1,1), (-1,1)], [-1,.3])  # rel tol 3e-14
            Gaussian Hidden Markov Model with 2 States
            Transition matrix:
            [                0.0                 1.0]
            [0.42857142857142855  0.5714285714285714]
            Emission parameters:
            [(1.0, 1.0), (-1.0, 1.0)]
            Initial probabilities: [0.0000, 1.0000]

        Bad things can happen::

            sage: hmm.GaussianHiddenMarkovModel([[-1,.7],[.3,.4]], [(1,1), (-1,1)], [-1,.3],
            ....:                               normalize=False)
            Gaussian Hidden Markov Model with 2 States
            Transition matrix:
            [-1.0  0.7]
            [ 0.3  0.4]
            ...
        """
        self.pi = util.initial_probs_to_TimeSeries(pi, normalize)
        self.N = len(self.pi)
        self.A = util.state_matrix_to_TimeSeries(A, self.N, normalize)

        # B should be a matrix of N rows, with column 0 the mean and 1
        # the standard deviation.
        if isinstance(B, Matrix):
            B = B.list()
        else:
            B = flatten(B)
        self.B = TimeSeries(B)
        self.probability_init()

    def __richcmp__(self, other, op):
        r"""
        Compare ``self`` and ``other``, which must both be GaussianHiddenMarkovModel's.

        EXAMPLES::

            sage: m = hmm.GaussianHiddenMarkovModel([[1]], [(0,1)], [1])
            sage: n = hmm.GaussianHiddenMarkovModel([[1]], [(1,1)], [1])
            sage: m < n
            True
            sage: m == m
            True
            sage: n > m
            True
            sage: n < m
            False
        """
        if not isinstance(other, GaussianHiddenMarkovModel):
            return NotImplemented
        return PyObject_RichCompare(self.__reduce__()[1],
                                    other.__reduce__()[1], op)

    def __getitem__(self, Py_ssize_t i):
        r"""
        Return the mean and standard distribution for the `i`-th state.

        INPUT:

        - ``i`` -- integer

        OUTPUT: 2 floats

        EXAMPLES::

            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]], [(1,.5), (-2,.3)], [.5,.5])
            sage: m[0]
            (1.0, 0.5)
            sage: m[1]
            (-2.0, 0.3)
            sage: m[-1]
            (-2.0, 0.3)
            sage: m[3]
            Traceback (most recent call last):
            ...
            IndexError: index out of range
            sage: m[-3]
            Traceback (most recent call last):
            ...
            IndexError: index out of range
        """
        if i < 0:
            i += self.N
        if i < 0 or i >= self.N:
            raise IndexError('index out of range')

        # TODO: change to be a normal distribution class (next version)
        return self.B[2*i], self.B[2*i+1]

    def __reduce__(self):
        r"""
        Used in pickling.

        EXAMPLES::

            sage: G = hmm.GaussianHiddenMarkovModel([[1]], [(0,1)], [1])
            sage: loads(dumps(G)) == G
            True
        """
        return unpickle_gaussian_hmm_v1, \
               (self.A, self.B, self.pi, self.prob, self.n_out)

    def emission_parameters(self):
        r"""
        Return the parameters that define the normal distributions
        associated to all of the states.

        OUTPUT:

        a list ``B`` of pairs ``B[i] = (mu, std)``, such that the
        distribution associated to state `i` is normal with mean
        ``mu`` and standard deviation ``std``.

        EXAMPLES::

            sage: M = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
            ....:                                   [(1,.5), (-1,3)],
            ....:                                   [.1,.9])
            sage: M.emission_parameters()
            [(1.0, 0.5), (-1.0, 3.0)]
        """
        cdef Py_ssize_t i
        from sage.rings.real_double import RDF
        return [(RDF(self.B[2*i]),RDF(self.B[2*i+1])) for i in range(self.N)]

    def __repr__(self):
        r"""
        Return string representation.

        EXAMPLES::

            sage: hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]], [(1,.5), (-1,3)], [.1,.9]).__repr__()
            'Gaussian Hidden Markov Model with 2 States\nTransition matrix:\n[0.1 0.9]\n[0.5 0.5]\nEmission parameters:\n[(1.0, 0.5), (-1.0, 3.0)]\nInitial probabilities: [0.1000, 0.9000]'
        """
        s = "Gaussian Hidden Markov Model with %s States" % self.N
        s += '\nTransition matrix:\n%s' % self.transition_matrix()
        s += '\nEmission parameters:\n%s' % self.emission_parameters()
        s += '\nInitial probabilities: %s' % self.initial_probabilities()
        return s

    def generate_sequence(self, Py_ssize_t length, starting_state=None):
        r"""
        Return a sample of the given length from this HMM.

        INPUT:

        - ``length`` -- positive integer
        - ``starting_state`` -- integer (or ``None``); if specified then
          generate a sequence using this model starting with the given state
          instead of the initial probabilities to determine the
          starting state.

        OUTPUT:

        - an :class:`IntList` or list of emission symbols
        - :class:`TimeSeries` of emissions

        EXAMPLES::

            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
            ....:                                   [(1,.5), (-1,3)],
            ....:                                   [.1,.9])
            sage: m.generate_sequence(5)  # random
            ([-3.0505, 0.5317, -4.5065, 0.6521, 1.0435], [1, 0, 1, 0, 1])
            sage: m.generate_sequence(0)
            ([], [])
            sage: m.generate_sequence(-1)
            Traceback (most recent call last):
            ...
            ValueError: length must be nonnegative

        Verify numerically that the starting state is 0 with probability about 0.1::

            sage: counter = 0
            sage: n = 0
            sage: def add_samples(i):
            ....:     global counter, n
            ....:     for i in range(i):
            ....:         n += 1
            ....:         if m.generate_sequence(1)[1][0] == 0:
            ....:             counter += 1

            sage: add_samples(10^5)
            sage: while abs(counter*1.0 / n - 0.1) > 0.01: add_samples(10^5)

        Example in which the starting state is 0 (see :issue:`11452`)::

            sage: set_random_seed(23);  m.generate_sequence(2)
            ([0.6501, -2.0151], [0, 1])

        Force a starting state of 1 even though as we saw above it would be 0::

            sage: set_random_seed(23);  m.generate_sequence(2, starting_state=1)
            ([-3.1491, -1.0244], [1, 1])
        """
        if length < 0:
            raise ValueError("length must be nonnegative")

        # Create Integer lists for states and observations
        cdef IntList states = IntList(length)
        cdef TimeSeries obs = TimeSeries(length)
        if length == 0:
            return states, obs

        # Setup variables, including random state.
        cdef Py_ssize_t i, j
        cdef randstate rstate = current_randstate()
        cdef int q = 0
        cdef double r, accum

        # Choose the starting state.
        # See the remark in hmm.pyx about how this should get
        # replaced by some general fast discrete distribution code.
        if starting_state is None:
            r = rstate.c_rand_double()
            accum = 0
            for i in range(self.N):
                if r < self.pi._values[i] + accum:
                    q = i
                    break
                else:
                    accum += self.pi._values[i]
        else:
            q = starting_state
            if q < 0 or q>= self.N:
                raise ValueError("starting state must be between 0 and %s" % (self.N-1))

        states._values[0] = q
        obs._values[0] = self.random_sample(q, rstate)

        cdef double* row
        cdef int O
        sig_on()
        for i in range(1, length):
            accum = 0
            row = self.A._values + q*self.N
            r = rstate.c_rand_double()
            for j in range(self.N):
                if r < row[j] + accum:
                    q = j
                    break
                else:
                    accum += row[j]
            states._values[i] = q
            obs._values[i] = self.random_sample(q, rstate)
        sig_off()

        return obs, states

    cdef probability_init(self):
        r"""
        Used internally to compute caching information that makes
        certain computations in the Baum-Welch algorithm faster.  This
        function has no input or output.
        """
        self.prob = TimeSeries(2*self.N)
        cdef int i
        for i in range(self.N):
            self.prob[2*i] = 1.0/(sqrt2pi*self.B[2*i+1])
            self.prob[2*i+1] = -1.0/(2*self.B[2*i+1]*self.B[2*i+1])

    cdef double random_sample(self, int state, randstate rstate) noexcept:
        r"""
        Return a random sample from the normal distribution associated
        to the given state.

        This is only used internally, and no bounds or other error
        checking is done, so calling this improperly can lead to seg
        faults.

        INPUT:

        - ``state`` -- integer
        - ``rstate`` -- randstate instance

        OUTPUT:

        double
        """
        return random_normal(self.B._values[state*2], self.B._values[state*2+1], rstate)

    cdef double probability_of(self, int state, double observation) noexcept:
        r"""
        Return a useful continuous analogue of "the probability b_j(o)"
        of seeing the given observation given that we're in the given
        state j (=state).

        The distribution is a normal distribution, and we're asking
        about the probability of a particular point being observed;
        the probability of a particular point is 0, which is not
        useful.  Thus we instead consider the limit p = prob([o,o+d])/d
        as d goes to 0.  There is a simple closed form formula for p,
        derived in the source code.  Note that p can be bigger than 1;
        for example, if we set observation=mean in the closed formula
        we get p=1/(sqrt(2*pi)*std), so p>1 when std<1/sqrt(2*pi).

        INPUT:

        - ``state`` -- integer
        - ``observation`` -- double

        OUTPUT:

        double
        """
        # The code below is an optimized version of the following code:
        #     cdef double mean = self.B._values[2*state], \
        #                 std = self.B._values[2*state+1]
        #     return 1/(sqrt2pi*std) * \
        #             exp(-(observation-mean)*(observation-mean)/(2*std*std))
        #
        # Here is how to use Sage to verify that the above formula computes
        # the limit claimed above:
        #
        # var('x,d,obs,mean,std')
        # n = 1/sqrt(2*pi*std^2) * exp(-(x-mean)^2/(2*std^2))
        # assume(std>0); assume(d>0)
        # m = n.integrate(x,obs,obs+d)/d
        # p = SR(m.limit(d=0).simplify_full())
        # q = 1/(sqrt(2*pi)*std) * exp(-(obs-mean)*(obs-mean)/(2*std*std))
        # bool(p==q)  # outputs True

        cdef double x = observation - self.B._values[2*state] # observation - mean
        return self.prob._values[2*state] * exp(x*x*self.prob._values[2*state+1])

    def log_likelihood(self, obs):
        r"""
        Return the logarithm of a continuous analogue of the
        probability that this model produced the given observation
        sequence.

        Note that the "continuous analogue of the probability" above can
        be bigger than 1, hence the logarithm can be positive.

        INPUT:

        - ``obs`` -- sequence of observations

        OUTPUT: float

        EXAMPLES::

            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
            ....:                                   [(1,.5), (-1,3)],
            ....:                                   [.1,.9])
            sage: m.log_likelihood([1,1,1])
            -4.297880766072486
            sage: s = m.sample(20)
            sage: -80 < m.log_likelihood(s) < -20
            True
        """
        if len(obs) == 0:
            return 1.0
        if not isinstance(obs, TimeSeries):
            obs = TimeSeries(obs)
        return self._forward_scale(obs)

    def _forward_scale(self, TimeSeries obs):
        r"""
        Memory-efficient implementation of the forward algorithm (with scaling).

        INPUT:

        - ``obs`` -- integer list of observation states

        OUTPUT:

        float -- the log of the probability that the model
        produced this sequence

        EXAMPLES::

            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]], [(1,.5), (-1,3)], [.1,.9])
            sage: m._forward_scale(stats.TimeSeries([1,-1,-1,1]))
            -7.641988207069133
        """
        cdef Py_ssize_t i, j, t, T = len(obs)

        # The running sum of the log probabilities
        cdef double log_probability = 0, sum, a

        cdef TimeSeries alpha  = TimeSeries(self.N), \
                        alpha2 = TimeSeries(self.N)

        # Initialization
        sum = 0
        for i in range(self.N):
            a = self.pi[i] * self.probability_of(i, obs._values[0])
            alpha[i] = a
            sum += a

        log_probability = log(sum)
        alpha.rescale(1/sum)

        # Induction
        cdef double s
        for t in range(1, T):
            sum = 0
            for j in range(self.N):
                s = 0
                for i in range(self.N):
                    s += alpha._values[i] * self.A._values[i*self.N + j]
                a = s * self.probability_of(j, obs._values[t])
                alpha2._values[j] = a
                sum += a

            log_probability += log(sum)
            for j in range(self.N):
                alpha._values[j] = alpha2._values[j] / sum

        # Termination
        return log_probability

    def viterbi(self, obs):
        r"""
        Determine "the" hidden sequence of states that is most likely
        to produce the given sequence ``obs`` of observations, along with
        the probability that this hidden sequence actually produced
        the observation.

        INPUT:

        - ``obs`` -- sequence of emitted ints or symbols

        OUTPUT:

        - ``list`` -- "the" most probable sequence of hidden states, i.e.,
          the Viterbi path

        - ``float`` -- log of probability that the observed sequence
          was produced by the Viterbi sequence of states

        EXAMPLES:

        We find the optimal state sequence for a given model::

            sage: m = hmm.GaussianHiddenMarkovModel([[0.5,0.5],[0.5,0.5]],
            ....:                                   [(0,1),(10,1)],
            ....:                                   [0.5,0.5])
            sage: m.viterbi([0,1,10,10,1])
            ([0, 0, 1, 1, 0], -9.0604285688230...)

        Another example in which the most likely states change based
        on the last observation::

            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
            ....:                                   [(1,.5), (-1,3)],
            ....:                                   [.1,.9])
            sage: m.viterbi([-2,-1,.1,0.1])
            ([1, 1, 0, 1], -9.61823698847639...)
            sage: m.viterbi([-2,-1,.1,0.3])
            ([1, 1, 1, 0], -9.566023653378513)
        """
        cdef TimeSeries _obs
        if not isinstance(obs, TimeSeries):
            _obs = TimeSeries(obs)
        else:
            _obs = obs

        # The algorithm is the same as _viterbi above, except
        # we take the logarithms of everything first, and add
        # instead of multiply.
        cdef Py_ssize_t t, T = _obs._length
        cdef IntList state_sequence = IntList(T)
        if T == 0:
            return state_sequence, 0.0

        cdef int i, j, N = self.N

        # delta[i] is the maximum of the probabilities over all
        # paths ending in state i.
        cdef TimeSeries delta = TimeSeries(N), delta_prev = TimeSeries(N)

        # We view psi as an N x T matrix of ints. The quantity
        #          psi[N*t + j]
        # is a most probable hidden state at time t-1, given
        # that j is a most probable state at time j.
        cdef IntList psi = IntList(N * T)  # initialized to 0 by default

        # Log Preprocessing
        cdef TimeSeries A = self.A.log()
        cdef TimeSeries pi = self.pi.log()

        # Initialization:
        for i in range(N):
            delta._values[i] = pi._values[i] + log(self.probability_of(i, _obs._values[0]))

        # Recursion:
        cdef double mx, tmp, minus_inf = float('-inf')
        cdef int index

        for t in range(1, T):
            delta_prev, delta = delta, delta_prev
            for j in range(N):
                # Compute delta_t[j] = max_i(delta_{t-1}(i) a_{i,j}) * b_j(_obs[t])
                mx = minus_inf
                index = -1
                for i in range(N):
                    tmp = delta_prev._values[i] + A._values[i*N+j]
                    if tmp > mx:
                        mx = tmp
                        index = i
                delta._values[j] = mx + log(self.probability_of(j, _obs._values[t]))
                psi._values[N*t + j] = index

        # Termination:
        mx, index = delta.max(index=True)

        # Path (state sequence) backtracking:
        state_sequence._values[T-1] = index
        t = T-2
        while t >= 0:
            state_sequence._values[t] = psi._values[N*(t+1) + state_sequence._values[t+1]]
            t -= 1

        return state_sequence, mx

    cdef TimeSeries _backward_scale_all(self, TimeSeries obs, TimeSeries scale):
        r"""
        This function returns the matrix beta_t(i), and is used
        internally as part of the Baum-Welch algorithm.

        The quantity beta_t(i) is the probability of observing the
        sequence obs[t+1:] assuming that the model is in state i at
        time t.

        INPUT:

        - ``obs`` -- TimeSeries
        - ``scale`` -- TimeSeries

        OUTPUT:

        - TimeSeries beta such that beta_t(i) = beta[t*N + i]
        - scale is also changed by this function
        """
        cdef Py_ssize_t t, T = obs._length
        cdef int N = self.N, i, j
        cdef double s
        cdef TimeSeries beta = TimeSeries(N*T, initialize=False)

        # 1. Initialization
        for i in range(N):
            beta._values[(T-1)*N + i] = 1 / scale._values[T-1]

        # 2. Induction
        t = T-2
        while t >= 0:
            for i in range(N):
                s = 0
                for j in range(N):
                    s += self.A._values[i*N+j] * \
                         self.probability_of(j, obs._values[t+1]) * beta._values[(t+1)*N+j]
                beta._values[t*N + i] = s/scale._values[t]
            t -= 1
        return beta

    cdef _forward_scale_all(self, TimeSeries obs):
        r"""
        Return scaled values alpha_t(i), the sequence of scalings, and
        the log probability.

        The quantity alpha_t(i) is the probability of observing the
        sequence obs[:t+1] assuming that the model is in state i at
        time t.

        INPUT:

        - ``obs`` -- TimeSeries

        OUTPUT:

        - TimeSeries alpha with alpha_t(i) = alpha[t*N + i]
        - TimeSeries scale with scale[t] the scaling at step t
        - ``float`` -- log_probability of the observation sequence
          being produced by the model
        """
        cdef Py_ssize_t i, j, t, T = len(obs)
        cdef int N = self.N

        # The running some of the log probabilities
        cdef double log_probability = 0, sum, a

        cdef TimeSeries alpha = TimeSeries(N*T, initialize=False)
        cdef TimeSeries scale = TimeSeries(T, initialize=False)

        # Initialization
        sum = 0
        for i in range(self.N):
            a = self.pi._values[i] * self.probability_of(i, obs._values[0])
            alpha._values[0*N + i] = a
            sum += a

        scale._values[0] = sum
        log_probability = log(sum)
        for i in range(self.N):
            alpha._values[0*N + i] /= sum

        # Induction
        # The code below is just an optimized version of:
        #     alpha = (alpha * A).pairwise_product(B[O[t+1]])
        #     alpha = alpha.scale(1/alpha.sum())
        # along with keeping track of the log of the scaling factor.
        cdef double s
        for t in range(1,T):
            sum = 0
            for j in range(N):
                s = 0
                for i in range(N):
                    s += alpha._values[(t-1)*N + i] * self.A._values[i*N + j]
                a = s * self.probability_of(j, obs._values[t])
                alpha._values[t*N + j] = a
                sum += a

            log_probability += log(sum)
            scale._values[t] = sum
            for j in range(self.N):
                alpha._values[t*N + j] /= sum

        # Termination
        return alpha, scale, log_probability

    cdef TimeSeries _baum_welch_xi(self, TimeSeries alpha, TimeSeries beta, TimeSeries obs):
        r"""
        Used internally to compute the scaled quantity xi_t(i,j)
        appearing in the Baum-Welch reestimation algorithm.

        INPUT:

        - ``alpha`` -- TimeSeries as output by the scaled forward algorithm
        - ``beta`` -- TimeSeries as output by the scaled backward algorithm
        - ``obs`` -- TimeSeries of observations

        OUTPUT:

        TimeSeries xi such that xi[t*N*N + i*N + j] = xi_t(i,j).
        """
        cdef int i, j, N = self.N
        cdef double sum
        cdef Py_ssize_t t, T = alpha._length//N
        cdef TimeSeries xi = TimeSeries(T*N*N, initialize=False)
        for t in range(T-1):
            sum = 0.0
            for i in range(N):
                for j in range(N):
                    xi._values[t*N*N+i*N+j] = alpha._values[t*N+i]*beta._values[(t+1)*N+j]*\
                       self.A._values[i*N+j] * self.probability_of(j, obs._values[t+1])
                    sum += xi._values[t*N*N+i*N+j]
            for i in range(N):
                for j in range(N):
                    xi._values[t*N*N+i*N+j] /= sum
        return xi

    def baum_welch(self, obs, int max_iter=500, double log_likelihood_cutoff=1e-4,
                   double min_sd=0.01, bint fix_emissions=False, bint v=False):
        r"""
        Given an observation sequence ``obs``, improve this HMM using the
        Baum-Welch algorithm to increase the probability of observing ``obs``.

        INPUT:

        - ``obs`` -- a time series of emissions

        - ``max_iter`` -- integer (default: 500); maximum number
          of Baum-Welch steps to take

        - ``log_likelihood_cutoff`` -- positive float (default: 1e-4);
          the minimal improvement in likelihood with respect to
          the last iteration required to continue. Relative value
          to log likelihood.

        - ``min_sd`` -- positive float (default: 0.01); when
          reestimating, the standard deviation of emissions is not
          allowed to be less than ``min_sd``.

        - ``fix_emissions`` -- boolean (default: ``False``); if ``True``, do not
          change emissions when updating

        OUTPUT:

        changes the model in place, and returns the log
        likelihood and number of iterations.

        EXAMPLES::

            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
            ....:                                   [(1,.5), (-1,3)],
            ....:                                   [.1,.9])
            sage: m.log_likelihood([-2,-1,.1,0.1])
            -8.858282215986275
            sage: m.baum_welch([-2,-1,.1,0.1])
            (4.534646052182..., 7)
            sage: m.log_likelihood([-2,-1,.1,0.1])
            4.534646052182...
            sage: m  # rel tol 3e-14
            Gaussian Hidden Markov Model with 2 States
            Transition matrix:
            [   0.9999999992430161 7.569839394440382e-10]
            [  0.49998462791192644    0.5000153720880736]
            Emission parameters:
            [(0.09999999999999999, 0.01), (-1.4999508147591902, 0.5000710504895474)]
            Initial probabilities: [0.0000, 1.0000]

        We illustrate bounding the standard deviation below.  Note that above we had
        different emission parameters when the ``min_sd`` was the default of 0.01::

            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
            ....:                                   [(1,.5), (-1,3)],
            ....:                                   [.1,.9])
            sage: m.baum_welch([-2,-1,.1,0.1], min_sd=1)
            (-4.07939572755..., 32)
            sage: m.emission_parameters()
            [(-0.2663018798..., 1.0), (-1.99850979..., 1.0)]

        We watch the log likelihoods of the model converge, step by
        step. We set the random seed beforehand so that the output is
        predictable (and guaranteed to avoid random floating point
        issues)::

            sage: set_random_seed(0)
            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
            ....:                                   [(1,.5), (-1,3)],
            ....:                                   [.1,.9])
            sage: v = m.sample(10)
            sage: l = stats.TimeSeries([m.baum_welch(v, max_iter=1)[0]
            ....:                       for _ in range(len(v))])
            sage: all(l[i] <= l[i+1] + 0.0001 for i in range(9))
            True
            sage: l
            [-20.1167, -17.7611, -16.9814, -16.9364, -16.9314,
             -16.9309, -16.9309, -16.9309, -16.9309, -16.9309]

        We illustrate fixing emissions; again the random seed is fixed
        to ensure that the output is what we expect it to be::

            sage: set_random_seed(0)
            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.9,.1]],
            ....:                                   [(1,2),(-1,.5)],
            ....:                                   [.3,.7])
            sage: v = m.sample(100)
            sage: m.baum_welch(v,fix_emissions=True)
            (-164.72944548204..., 23)
            sage: m.emission_parameters()
            [(1.0, 2.0), (-1.0, 0.5)]
            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.9,.1]],
            ....:                                   [(1,2),(-1,.5)],
            ....:                                   [.3,.7])
            sage: m.baum_welch(v)
            (-162.854370397998..., 49)
            sage: m.emission_parameters()  # rel tol 3e-14
            [(1.2722419172602375, 2.371368751761901),
             (-0.9486174675179113, 0.5762360385123765)]

        TESTS:

        If your luck is bad enough to cause a :exc:`RuntimeError`,
        re-sampling will fix it. This is a repeat of one of the
        examples above, without the fixed random seed, but then
        acknowledging that the algorithm can fail in rare cases::

            sage: m = hmm.GaussianHiddenMarkovModel([[.1,.9],[.5,.5]],
            ....:                                   [(1,.5), (-1,3)],
            ....:                                   [.1,.9])
            sage: v = m.sample(10)
            sage: while True:
            ....:     try:
            ....:         l = stats.TimeSeries([m.baum_welch(v, max_iter=1)[0]
            ....:                               for _ in range(len(v))])
            ....:         break
            ....:     except RuntimeError:
            ....:         v = m.sample(10)
            sage: all(l[i] <= l[i+1] + 0.0001 for i in range(9))
            True

        """
        if not isinstance(obs, TimeSeries):
            obs = TimeSeries(obs)
        cdef TimeSeries _obs = obs
        cdef TimeSeries alpha, beta, scale, gamma, xi
        cdef double log_probability, log_probability0, log_probability_prev, delta
        cdef int i, j, k, N, n_iterations
        cdef Py_ssize_t t, T
        cdef double denominator_A, numerator_A, denominator_B, numerator_mean, numerator_std

        # Initialization
        alpha, scale, log_probability0 = self._forward_scale_all(_obs)
        if not isfinite(log_probability0):
            return (0.0, 0)
        log_probability = log_probability0
        beta = self._backward_scale_all(_obs, scale)
        gamma = self._baum_welch_gamma(alpha, beta)
        xi = self._baum_welch_xi(alpha, beta, _obs)
        log_probability_prev = log_probability
        N = self.N
        n_iterations = 0
        T = len(_obs)

        # Re-estimation
        while True:
            # Reestimate
            for i in range(N):
                if not isfinite(gamma._values[0*N+i]):
                    # Before raising an error, leave self in a valid state.
                    util.normalize_probability_TimeSeries(self.pi, 0, self.pi._length)
                    raise RuntimeError("impossible to compute gamma during reestimation")
                self.pi._values[i] = gamma._values[0*N+i]

            # Update the probabilities pi to define a valid discrete distribution
            util.normalize_probability_TimeSeries(self.pi, 0, self.pi._length)

            # Reestimate transition matrix and emission probabilities in
            # each state.
            for i in range(N):
                # Compute the updated transition matrix
                denominator_A = 0.0
                for t in range(T-1):
                    denominator_A += gamma._values[t*N+i]
                if not isnormal(denominator_A):
                    raise RuntimeError("unable to re-estimate transition matrix")
                for j in range(N):
                    numerator_A = 0.0
                    for t in range(T-1):
                        numerator_A += xi._values[t*N*N+i*N+j]
                    self.A._values[i*N+j] = numerator_A / denominator_A

                # Rescale the i-th row of the transition matrix to be
                # a valid stochastic matrix:
                util.normalize_probability_TimeSeries(self.A, i*N, (i+1)*N)

                if not fix_emissions:
                    denominator_B = denominator_A + gamma._values[(T-1)*N + i]
                    if not isnormal(denominator_B):
                        raise RuntimeError("unable to re-estimate emission probabilities")

                    numerator_mean = 0.0
                    numerator_std = 0.0
                    for t in range(T):
                        numerator_mean += gamma._values[t*N + i] * _obs._values[t]
                        numerator_std  += gamma._values[t*N + i] * \
                               (_obs._values[t] - self.B._values[2*i])*(_obs._values[t] - self.B._values[2*i])
                    # restimated mean
                    self.B._values[2*i] = numerator_mean / denominator_B
                    # restimated standard deviation
                    self.B._values[2*i+1] = sqrt(numerator_std / denominator_B)
                    if self.B._values[2*i+1] < min_sd:
                        self.B._values[2*i+1] = min_sd
                    self.probability_init()

            n_iterations += 1
            if n_iterations >= max_iter: break

            # Initialization for next iteration
            alpha, scale, log_probability0 = self._forward_scale_all(_obs)

            if not isfinite(log_probability0): break
            log_probability = log_probability0
            beta = self._backward_scale_all(_obs, scale)
            gamma = self._baum_welch_gamma(alpha, beta)
            xi = self._baum_welch_xi(alpha, beta, _obs)

            # Compute the difference between the log probability of
            # two iterations.
            delta = log_probability - log_probability_prev
            log_probability_prev = log_probability

            # If the log probability does not change by more than delta,
            # then terminate
            if delta >= 0 and delta <= log_likelihood_cutoff:
                break

        return log_probability, n_iterations


cdef class GaussianMixtureHiddenMarkovModel(GaussianHiddenMarkovModel):
    r"""
    Gaussian mixture Hidden Markov Model.

    INPUT:

    - ``A`` -- matrix; the `N \times N` transition matrix

    - ``B`` -- list of mixture definitions for each state.  Each
      state may have a varying number of gaussians with selection
      probabilities that sum to 1 and encoded as ``(p, (mu,sigma))``

    - ``pi`` -- initial state probabilities

    - ``normalize`` -- boolean (default: ``True``); if given, input is
      normalized to define valid probability distributions,
      e.g., the entries of `A` are made nonnegative and the rows
      sum to 1, and the probabilities in ``pi`` are normalized.

    EXAMPLES::

        sage: A = [[0.5,0.5],[0.5,0.5]]
        sage: B = [[(0.9,(0.0,1.0)), (0.1,(1,10000))],[(1,(1,1)), (0,(0,0.1))]]
        sage: hmm.GaussianMixtureHiddenMarkovModel(A, B, [1,0])
        Gaussian Mixture Hidden Markov Model with 2 States
        Transition matrix:
        [0.5 0.5]
        [0.5 0.5]
        Emission parameters:
        [0.9*N(0.0,1.0) + 0.1*N(1.0,10000.0), 1.0*N(1.0,1.0) + 0.0*N(0.0,0.1)]
        Initial probabilities: [1.0000, 0.0000]

    TESTS:

    If a standard deviation is 0, it is normalized to be slightly bigger than 0.::

        sage: hmm.GaussianMixtureHiddenMarkovModel([[1]], [[(1,(0,0))]], [1])
        Gaussian Mixture Hidden Markov Model with 1 States
        Transition matrix:
        [1.0]
        Emission parameters:
        [1.0*N(0.0,1e-08)]
        Initial probabilities: [1.0000]

    We test that number of emission distributions must be the same as the number of states::

        sage: hmm.GaussianMixtureHiddenMarkovModel([[1]], [], [1])
        Traceback (most recent call last):
        ...
        ValueError: number of GaussianMixtures must be the same as number of entries of pi

        sage: hmm.GaussianMixtureHiddenMarkovModel([[1]], [[]], [1])
        Traceback (most recent call last):
        ...
        ValueError: must specify at least one component of the mixture model

    We test that the number of initial probabilities must equal the number of states::

        sage: hmm.GaussianMixtureHiddenMarkovModel([[1]], [[]], [1,2])
        Traceback (most recent call last):
        ...
        ValueError: number of entries of transition matrix A must be the square of the number of entries of pi
    """

    cdef object mixture # mixture

    def __init__(self, A, B, pi=None, bint normalize=True):
        r"""
        Initialize a Gaussian mixture hidden Markov model.

        EXAMPLES::

            sage: hmm.GaussianMixtureHiddenMarkovModel([[.9,.1],[.4,.6]], [[(.4,(0,1)), (.6,(1,0.1))],[(1,(0,1))]], [.7,.3])
            Gaussian Mixture Hidden Markov Model with 2 States
            Transition matrix:
            [0.9 0.1]
            [0.4 0.6]
            Emission parameters:
            [0.4*N(0.0,1.0) + 0.6*N(1.0,0.1), 1.0*N(0.0,1.0)]
            Initial probabilities: [0.7000, 0.3000]
        """
        self.pi = util.initial_probs_to_TimeSeries(pi, normalize)
        self.N = len(self.pi)
        self.A = util.state_matrix_to_TimeSeries(A, self.N, normalize)
        if self.N*self.N != len(self.A):
            raise ValueError("number of entries of transition matrix A must be the square of the number of entries of pi")

        self.mixture = [b if isinstance(b, GaussianMixtureDistribution) else
                        GaussianMixtureDistribution([flatten(x) for x in b])
                        for b in B]
        if len(self.mixture) != self.N:
            raise ValueError("number of GaussianMixtures must be the same as number of entries of pi")

    def __repr__(self):
        r"""
        Return string representation.

        EXAMPLES::

            sage: hmm.GaussianMixtureHiddenMarkovModel([[.9,.1],[.4,.6]], [[(.4,(0,1)), (.6,(1,0.1))],[(1,(0,1))]], [.7,.3]).__repr__()
            'Gaussian Mixture Hidden Markov Model with 2 States\nTransition matrix:\n[0.9 0.1]\n[0.4 0.6]\nEmission parameters:\n[0.4*N(0.0,1.0) + 0.6*N(1.0,0.1), 1.0*N(0.0,1.0)]\nInitial probabilities: [0.7000, 0.3000]'
        """
        s = "Gaussian Mixture Hidden Markov Model with %s States" % self.N
        s += '\nTransition matrix:\n%s' % self.transition_matrix()
        s += '\nEmission parameters:\n%s' % self.emission_parameters()
        s += '\nInitial probabilities: %s' % self.initial_probabilities()
        return s

    def __reduce__(self):
        r"""
        Used in pickling.

        EXAMPLES::

            sage: m = hmm.GaussianMixtureHiddenMarkovModel([[1]], [[(.4,(0,1)), (.6,(1,0.1))]], [1])
            sage: loads(dumps(m)) == m
            True
        """
        return unpickle_gaussian_mixture_hmm_v1, \
               (self.A, self.B, self.pi, self.mixture)

    def __richcmp__(self, other, op):
        r"""
        Compare ``self`` and ``other``, which must both be
        ``GaussianMixtureHiddenMarkovModel``s.

        EXAMPLES::

            sage: m = hmm.GaussianMixtureHiddenMarkovModel([[1]], [[(.4,(0,1)), (.6,(1,0.1))]], [1])
            sage: n = hmm.GaussianMixtureHiddenMarkovModel([[1]], [[(.5,(0,1)), (.5,(1,0.1))]], [1])
            sage: m < n
            True
            sage: m == m
            True
            sage: n > m
            True
            sage: n < m
            False
        """
        if not isinstance(other, GaussianMixtureHiddenMarkovModel):
            return NotImplemented
        return PyObject_RichCompare(self.__reduce__()[1],
                                    other.__reduce__()[1], op)

    def __getitem__(self, Py_ssize_t i):
        r"""
        Return the Gaussian mixture distribution associated to the
        i-th state.

        INPUT:

        - ``i`` -- integer

        OUTPUT: a Gaussian mixture distribution object

        EXAMPLES::

            sage: m = hmm.GaussianMixtureHiddenMarkovModel([[.9,.1],[.4,.6]], [[(.4,(0,1)), (.6,(1,0.1))],[(1,(0,1))]], [.7,.3])
            sage: m[0]
            0.4*N(0.0,1.0) + 0.6*N(1.0,0.1)
            sage: m[1]
            1.0*N(0.0,1.0)

        Negative indexing works::

            sage: m[-1]
            1.0*N(0.0,1.0)

        Bounds are checked::

            sage: m[2]
            Traceback (most recent call last):
            ...
            IndexError: index out of range
            sage: m[-3]
            Traceback (most recent call last):
            ...
            IndexError: index out of range
        """
        if i < 0:
            i += self.N
        if i < 0 or i >= self.N:
            raise IndexError('index out of range')
        return self.mixture[i]

    def emission_parameters(self):
        r"""
        Return a list of all the emission distributions.

        OUTPUT: list of Gaussian mixtures

        EXAMPLES::

            sage: m = hmm.GaussianMixtureHiddenMarkovModel([[.9,.1],[.4,.6]],
            ....:                                          [[(.4,(0,1)), (.6,(1,0.1))], [(1,(0,1))]],
            ....:                                          [.7,.3])
            sage: m.emission_parameters()
            [0.4*N(0.0,1.0) + 0.6*N(1.0,0.1), 1.0*N(0.0,1.0)]
        """
        return list(self.mixture)

    cdef double random_sample(self, int state, randstate rstate) noexcept:
        r"""
        Return a random sample from the normal distribution associated
        to the given state.

        This is only used internally, and no bounds or other error
        checking is done, so calling this improperly can lead to seg
        faults.

        INPUT:

        - ``state`` -- integer
        - ``rstate`` -- randstate instance

        OUTPUT:

        double
        """
        cdef GaussianMixtureDistribution G = self.mixture[state]
        return G._sample(rstate)

    cdef double probability_of(self, int state, double observation) noexcept:
        r"""
        Return the probability b_j(o) of see the given observation o
        (=observation) given that we're in the given state j (=state).

        This is a continuous probability, so this really returns a
        number p such that the probability of a value in the interval
        [o,o+d] is p*d.

        INPUT:

        - ``state`` -- integer
        - ``observation`` -- double

        OUTPUT:

        double
        """
        cdef GaussianMixtureDistribution G = self.mixture[state]
        return G.prob(observation)

    cdef TimeSeries _baum_welch_mixed_gamma(self, TimeSeries alpha, TimeSeries beta,
                                            TimeSeries obs, int j):
        r"""
        Let gamma_t(j,m) be the m-component (in the mixture) of the
        probability of being in state j at time t, given the
        observation sequence.  This function outputs a TimeSeries v
        such that v[m*T + t] gives gamma_t(j, m) where T is the number
        of time steps.

        INPUT:

        - ``alpha`` -- TimeSeries
        - ``beta`` -- TimeSeries
        - ``obs`` -- TimeSeries
        - ``j`` -- integer

        OUTPUT:

        TimeSeries
        """
        cdef int i, k, m, N = self.N
        cdef Py_ssize_t t, T = alpha._length//N

        cdef double numer, alpha_minus, P, s, prob
        cdef GaussianMixtureDistribution G = self.mixture[j]
        cdef int M = len(G)
        cdef TimeSeries mixed_gamma = TimeSeries(T*M)

        for t in range(T):
            prob = self.probability_of(j, obs._values[t])
            if prob == 0:
                # If the probability of observing obs[t] in state j is 0, then
                # all of the m-mixture components *have* to automatically be 0,
                # since prob is the sum of those and they are all nonnegative.
                for m in range(M):
                    mixed_gamma._values[m*T + t] = 0
            else:
                # Compute the denominator we used when scaling gamma.
                # The key thing is that this is consistent between
                # gamma and mixed_gamma.
                P = 0
                for k in range(N):
                    P += alpha._values[t*N+k]*beta._values[t*N+k]

                # Divide out the total probability, so we can multiply back in
                # the m-components of the probability.
                alpha_minus = alpha._values[t*N + j] / prob
                for m in range(M):
                    numer = alpha_minus * G.prob_m(obs._values[t], m) * beta._values[t*N + j]
                    mixed_gamma._values[m*T + t] = numer / P

        return mixed_gamma

    def baum_welch(self, obs, int max_iter=1000, double log_likelihood_cutoff=1e-12,
                   double min_sd=0.01, bint fix_emissions=False):
        r"""
        Given an observation sequence ``obs``, improve this HMM using the
        Baum-Welch algorithm to increase the probability of observing ``obs``.

        INPUT:

        - ``obs`` -- a time series of emissions
        - ``max_iter`` -- integer (default: 1000); maximum number
          of Baum-Welch steps to take
        - ``log_likelihood_cutoff`` -- positive float (default: 1e-12);
          the minimal improvement in likelihood with respect to
          the last iteration required to continue. Relative value
          to log likelihood.
        - ``min_sd`` -- positive float (default: 0.01); when
          reestimating, the standard deviation of emissions is not
          allowed to be less than ``min_sd``.
        - ``fix_emissions`` -- boolean (default: ``False``); if ``True``, do not
          change emissions when updating

        OUTPUT:

        changes the model in place, and returns the log
        likelihood and number of iterations.

        EXAMPLES::

            sage: m = hmm.GaussianMixtureHiddenMarkovModel(
            ....:         [[.9,.1],[.4,.6]],
            ....:         [[(.4,(0,1)), (.6,(1,0.1))], [(1,(0,1))]],
            ....:         [.7,.3])
            sage: set_random_seed(0); v = m.sample(10); v
            [0.3576, -0.9365, 0.9449, -0.6957, 1.0217,
             0.9644, 0.9987, -0.5950, -1.0219, 0.6477]
            sage: m.log_likelihood(v)
            -8.31408655939536...
            sage: m.baum_welch(v)
            (2.18905068682..., 15)
            sage: m.log_likelihood(v)
            2.18905068682...
            sage: m  # rel tol 6e-12
            Gaussian Mixture Hidden Markov Model with 2 States
            Transition matrix:
            [   0.8746363339773399   0.12536366602266016]
            [                  1.0 1.451685202290174e-40]
            Emission parameters:
            [0.500161629343*N(-0.812298726239,0.173329026744)
              + 0.499838370657*N(0.982433690378,0.029719932009),
             1.0*N(0.503260056832,0.145881515324)]
            Initial probabilities: [0.0000, 1.0000]

        We illustrate bounding the standard deviation below.  Note that above we had
        different emission parameters when the min_sd was the default of 0.01::

            sage: m = hmm.GaussianMixtureHiddenMarkovModel(
            ....:         [[.9,.1],[.4,.6]],
            ....:         [[(.4,(0,1)), (.6,(1,0.1))], [(1,(0,1))]],
            ....:         [.7,.3])
            sage: m.baum_welch(v, min_sd=1)
            (-12.617885761692..., 1000)
            sage: m.emission_parameters()  # rel tol 6e-12
            [0.503545634447*N(0.200166509595,1.0) + 0.496454365553*N(0.200166509595,1.0),
             1.0*N(0.0543433426535,1.0)]

        We illustrate fixing all emissions::

            sage: m = hmm.GaussianMixtureHiddenMarkovModel(
            ....:         [[.9,.1],[.4,.6]],
            ....:         [[(.4,(0,1)), (.6,(1,0.1))], [(1,(0,1))]],
            ....:         [.7,.3])
            sage: set_random_seed(0); v = m.sample(10)
            sage: m.baum_welch(v, fix_emissions=True)
            (-7.58656858997..., 36)
            sage: m.emission_parameters()
            [0.4*N(0.0,1.0) + 0.6*N(1.0,0.1),
             1.0*N(0.0,1.0)]
        """
        if not isinstance(obs, TimeSeries):
            obs = TimeSeries(obs)
        cdef TimeSeries _obs = obs
        cdef TimeSeries alpha, beta, scale, gamma, mixed_gamma, mixed_gamma_m, xi
        cdef double log_probability, log_probability0, log_probability_prev, delta
        cdef int i, j, k, m, N, n_iterations
        cdef Py_ssize_t t, T
        cdef double denominator_A, numerator_A, denominator_B, numerator_mean, numerator_std, \
             numerator_c, c, mu, std, numer, denom, new_mu, new_std, new_c, s
        cdef GaussianMixtureDistribution G

        # Initialization
        alpha, scale, log_probability0 = self._forward_scale_all(_obs)
        if not isfinite(log_probability0):
            return (0.0, 0)
        log_probability = log_probability0
        beta = self._backward_scale_all(_obs, scale)
        gamma = self._baum_welch_gamma(alpha, beta)
        xi = self._baum_welch_xi(alpha, beta, _obs)
        log_probability_prev = log_probability
        N = self.N
        n_iterations = 0
        T = len(_obs)

        # Re-estimation
        while True:

            # Reestimate frequency of state i in time t=0
            for i in range(N):
                if not isfinite(gamma._values[0*N+i]):
                    # Before raising an error, leave self in a valid state.
                    util.normalize_probability_TimeSeries(self.pi, 0, self.pi._length)
                    raise RuntimeError("impossible to compute gamma during reestimation")
                self.pi._values[i] = gamma._values[0*N+i]

            # Update the probabilities pi to define a valid discrete distribution
            util.normalize_probability_TimeSeries(self.pi, 0, self.pi._length)

            # Reestimate transition matrix and emission probabilities in
            # each state.
            for i in range(N):
                # Reestimate the state transition matrix
                denominator_A = 0.0
                for t in range(T-1):
                    denominator_A += gamma._values[t*N+i]
                if not isnormal(denominator_A):
                    raise RuntimeError("unable to re-estimate pi (1)")
                for j in range(N):
                    numerator_A = 0.0
                    for t in range(T-1):
                        numerator_A += xi._values[t*N*N+i*N+j]
                    self.A._values[i*N+j] = numerator_A / denominator_A

                # Rescale the i-th row of the transition matrix to be
                # a valid stochastic matrix:
                util.normalize_probability_TimeSeries(self.A, i*N, (i+1)*N)

                ########################################################################
                # Re-estimate the emission probabilities
                ########################################################################
                G = self.mixture[i]
                if not fix_emissions and not G.is_fixed():
                    mixed_gamma = self._baum_welch_mixed_gamma(alpha, beta, _obs, i)
                    new_G = []
                    for m in range(len(G)):
                        if G.fixed._values[m]:
                            new_G.append(G[m])
                            continue

                        # Compute re-estimated mu_{j,m}
                        numer = 0
                        denom = 0
                        for t in range(T):
                            numer += mixed_gamma._values[m*T + t] * _obs._values[t]
                            denom += mixed_gamma._values[m*T + t]
                        new_mu = numer / denom

                        # Compute re-estimated standard deviation
                        numer = 0
                        mu = G[m][1]
                        for t in range(T):
                            numer += mixed_gamma._values[m*T + t] * \
                                     (_obs._values[t] - mu)*(_obs._values[t] - mu)

                        new_std = sqrt(numer / denom)
                        if new_std < min_sd:
                            new_std = min_sd

                        # Compute re-estimated weighting coefficient
                        new_c = denom
                        s = 0
                        for t in range(T):
                            s += gamma._values[t*N + i]
                        new_c /= s

                        new_G.append((new_c,new_mu,new_std))

                    self.mixture[i] = GaussianMixtureDistribution(new_G)

            n_iterations += 1
            if n_iterations >= max_iter: break

            ########################################################################
            # Initialization for next iteration
            ########################################################################
            alpha, scale, log_probability0 = self._forward_scale_all(_obs)
            if not isfinite(log_probability0): break
            log_probability = log_probability0
            beta = self._backward_scale_all(_obs, scale)
            gamma = self._baum_welch_gamma(alpha, beta)
            xi = self._baum_welch_xi(alpha, beta, _obs)

            # Compute the difference between the log probability of
            # two iterations.
            delta = log_probability - log_probability_prev
            log_probability_prev = log_probability

            # If the log probability does not improve by more than
            # delta, then terminate
            if delta >= 0 and delta <= log_likelihood_cutoff:
                break

        return log_probability, n_iterations


##################################################
# For Unpickling
##################################################

# We keep the _v0 function for backwards compatible.
def unpickle_gaussian_hmm_v0(A, B, pi, name):
    r"""
    EXAMPLES::

        sage: m = hmm.GaussianHiddenMarkovModel([[1]], [(0,1)], [1])
        sage: sage.stats.hmm.chmm.unpickle_gaussian_hmm_v0(m.transition_matrix(), m.emission_parameters(), m.initial_probabilities(), 'test')
        Gaussian Hidden Markov Model with 1 States
        Transition matrix:
        [1.0]
        Emission parameters:
        [(0.0, 1.0)]
        Initial probabilities: [1.0000]
    """
    return GaussianHiddenMarkovModel(A,B,pi)


def unpickle_gaussian_hmm_v1(A, B, pi, prob, n_out):
    r"""
    EXAMPLES::

        sage: m = hmm.GaussianHiddenMarkovModel([[1]], [(0,1)], [1])
        sage: loads(dumps(m)) == m   # indirect test
        True
    """
    cdef GaussianHiddenMarkovModel m = GaussianHiddenMarkovModel.__new__(GaussianHiddenMarkovModel)
    m.A = A
    m.B = B
    m.pi = pi
    m.prob = prob
    m.n_out = n_out
    return m


def unpickle_gaussian_mixture_hmm_v1(A, B, pi, mixture):
    r"""
    EXAMPLES::

        sage: m = hmm.GaussianMixtureHiddenMarkovModel([[1]], [[(.4,(0,1)), (.6,(1,0.1))]], [1])
        sage: loads(dumps(m)) == m   # indirect test
        True
    """
    cdef GaussianMixtureHiddenMarkovModel m = GaussianMixtureHiddenMarkovModel.__new__(GaussianMixtureHiddenMarkovModel)
    m.A = A
    m.B = B
    m.pi = pi
    m.mixture = mixture
    return m



# sage.doctest: needs sage.modules sage.rings.finite_rings
r"""
Access functions to online databases for coding theory
"""
from sage.misc.lazy_import import lazy_import

# Import the following function so that it is available as
# - sage.coding.databases.self_dual_binary_codes
# - codes.databases.self_dual_binary_codes (which functions as a catalog).
lazy_import('sage.coding.self_dual_codes', 'self_dual_binary_codes')

del lazy_import

# Do not put any global imports here since this module is accessible as
# - sage.coding.databases.<tab>
# - codes.databases.<tab>


def best_linear_code_in_guava(n, k, F):
    r"""
    Return the linear code of length ``n``, dimension ``k`` over field ``F``
    with the maximal minimum distance which is known to the GAP package GUAVA.

    The function uses the tables described in :func:`bounds_on_minimum_distance_in_guava` to
    construct this code. This requires the optional GAP package GUAVA.

    INPUT:

    - ``n`` -- the length of the code to look up

    - ``k`` -- the dimension of the code to look up

    - ``F`` -- the base field of the code to look up

    OUTPUT: a :class:`LinearCode` which is a best linear code of the given parameters known to GUAVA

    EXAMPLES::

        sage: codes.databases.best_linear_code_in_guava(10,5,GF(2))    # long time; optional - gap_package_guava
        [10, 5] linear code over GF(2)
        sage: libgap.LoadPackage('guava')                              # long time; optional - gap_package_guava
        ...
        sage: libgap.BestKnownLinearCode(10,5,libgap.GF(2))            # long time; optional - gap_package_guava
        a linear [10,5,4]2..4 shortened code

    This means that the best possible binary linear code of length 10 and
    dimension 5 is a code with minimum distance 4 and covering radius s somewhere
    between 2 and 4. Use ``bounds_on_minimum_distance_in_guava(10,5,GF(2))``
    for further details.
    """
    from sage.features.gap import GapPackage
    from .linear_code import LinearCode
    GapPackage('guava', spkg='gap_packages').require()
    from sage.libs.gap.libgap import libgap
    libgap.load_package('guava')
    C = libgap.BestKnownLinearCode(n, k, F)
    return LinearCode(C.GeneratorMat()._matrix_(F))


def bounds_on_minimum_distance_in_guava(n, k, F):
    r"""
    Compute a lower and upper bound on the greatest minimum distance of a
    `[n,k]` linear code over the field ``F``.

    This function requires the optional GAP package GUAVA.

    The function returns a GAP record with the two bounds and an explanation for
    each bound. The method ``Display`` can be used to show the explanations.

    The values for the lower and upper bound are obtained from a table
    constructed by Cen Tjhai for GUAVA, derived from the table of
    Brouwer. See http://www.codetables.de/ for the most recent data.
    These tables contain lower and upper bounds for `q=2` (when ``n <= 257``),
    `q=3` (when ``n <= 243``), `q=4` (``n <= 256``). (Current as of
    11 May 2006.) For codes over other fields and for larger word lengths,
    trivial bounds are used.

    INPUT:

    - ``n`` -- the length of the code to look up

    - ``k`` -- the dimension of the code to look up

    - ``F`` -- the base field of the code to look up

    OUTPUT: a GAP record object. See below for an example

    EXAMPLES::

        sage: gap_rec = codes.databases.bounds_on_minimum_distance_in_guava(10,5,GF(2))  # optional - gap_package_guava
        sage: gap_rec.Display()                                                          # optional - gap_package_guava
        rec(
          construction := [ <Operation "ShortenedCode">,
            [ [ <Operation "UUVCode">,
              [ [ <Operation "DualCode">,
              [ [ <Operation "RepetitionCode">, [ 8, 2 ] ] ] ],
              [ <Operation "UUVCode">, [ [ <Operation "DualCode">,
                [ [ <Operation "RepetitionCode">, [ 4, 2 ] ] ] ],
                [ <Operation "RepetitionCode">, [ 4, 2 ] ] ] ] ] ],
            [ 1, 2, 3, 4, 5, 6 ] ] ],
          k := 5,
          lowerBound := 4,
          lowerBoundExplanation := ...
          n := 10,
          q := 2,
          references := rec(
               ),
          upperBound := 4,
          upperBoundExplanation := ... )
    """
    from sage.features.gap import GapPackage
    GapPackage('guava', spkg='gap_packages').require()
    from sage.libs.gap.libgap import libgap
    libgap.load_package('guava')
    return libgap.BoundsMinimumDistance(n, k, F)


def best_linear_code_in_codetables_dot_de(n, k, F, verbose=False):
    r"""
    Return the best linear code and its construction as per the web database
    http://www.codetables.de/

    INPUT:

    - ``n`` -- integer; the length of the code

    - ``k`` -- integer; the dimension of the code

    - ``F`` -- finite field, of order 2, 3, 4, 5, 7, 8, or 9

    - ``verbose`` -- boolean (default: ``False``)

    OUTPUT: an unparsed text explaining the construction of the code

    EXAMPLES::

        sage: L = codes.databases.best_linear_code_in_codetables_dot_de(72, 36, GF(2))    # optional - internet
        sage: print(L)                                                                    # optional - internet
        Construction of a linear code
        [72,36,15] over GF(2):
        [1]:  [73, 36, 16] Cyclic Linear Code over GF(2)
             CyclicCode of length 73 with generating polynomial x^37 + x^36 + x^34 +
        x^33 + x^32 + x^27 + x^25 + x^24 + x^22 + x^21 + x^19 + x^18 + x^15 + x^11 +
        x^10 + x^8 + x^7 + x^5 + x^3 + 1
        [2]:  [72, 36, 15] Linear Code over GF(2)
             Puncturing of [1] at 1
        <BLANKLINE>
        last modified: 2002-03-20

    This function raises an :exc:`IOError` if an error occurs downloading data or
    parsing it. It raises a :exc:`ValueError` if the ``q`` input is invalid.

    AUTHORS:

    - Steven Sivek (2005-11-14)
    - David Joyner (2008-03)
    """
    from urllib.request import urlopen
    from sage.cpython.string import bytes_to_str
    q = F.order()
    if q not in [2, 3, 4, 5, 7, 8, 9]:
        raise ValueError("q (=%s) must be in [2,3,4,5,7,8,9]" % q)
    n = int(n)
    k = int(k)

    param = ("?q=%s&n=%s&k=%s" % (q, n, k)).replace('L', '')

    url = "http://www.codetables.de/" + "BKLC/BKLC.php" + param
    if verbose:
        print("Looking up the bounds at %s" % url)
    with urlopen(url) as f:
        s = f.read()

    s = bytes_to_str(s)
    i = s.find("<PRE>")
    j = s.find("</PRE>")
    if i == -1 or j == -1:
        raise OSError("Error parsing data (missing pre tags).")
    return s[i+5:j].strip()


def self_orthogonal_binary_codes(n, k, b=2, parent=None, BC=None, equal=False,
    in_test=None):
    """
    Return a Python iterator which generates a complete set of
    representatives of all permutation equivalence classes of
    self-orthogonal binary linear codes of length in ``[1..n]`` and
    dimension in ``[1..k]``.

    INPUT:

    - ``n`` -- integer; maximal length

    - ``k`` -- integer; maximal dimension

    - ``b`` -- integer; requires that the generators all have weight divisible
      by ``b`` (if ``b=2``, all self-orthogonal codes are generated, and if
      ``b=4``, all doubly even codes are generated). Must be an even positive
      integer.

    - ``parent`` -- used in recursion (default: ``None``)

    - ``BC`` -- used in recursion (default: ``None``)

    - ``equal`` -- if ``True``, generates only [n, k] codes (default: ``False``)

    - ``in_test`` -- used in recursion (default: ``None``)

    EXAMPLES:

    Generate all self-orthogonal codes of length up to 7 and dimension up
    to 3::

        sage: # needs sage.groups
        sage: for B in codes.databases.self_orthogonal_binary_codes(7,3):
        ....:    print(B)
        [2, 1] linear code over GF(2)
        [4, 2] linear code over GF(2)
        [6, 3] linear code over GF(2)
        [4, 1] linear code over GF(2)
        [6, 2] linear code over GF(2)
        [6, 2] linear code over GF(2)
        [7, 3] linear code over GF(2)
        [6, 1] linear code over GF(2)

    Generate all doubly-even codes of length up to 7 and dimension up
    to 3::

        sage: # needs sage.groups
        sage: for B in codes.databases.self_orthogonal_binary_codes(7,3,4):
        ....:    print(B); print(B.generator_matrix())
        [4, 1] linear code over GF(2)
        [1 1 1 1]
        [6, 2] linear code over GF(2)
        [1 0 1 0 1 1]
        [0 1 0 1 1 1]
        [7, 3] linear code over GF(2)
        [1 0 0 1 1 0 1]
        [0 1 0 1 1 1 0]
        [0 0 1 0 1 1 1]

    Generate all doubly-even codes of length up to 7 and dimension up
    to 2::

        sage: # needs sage.groups
        sage: for B in codes.databases.self_orthogonal_binary_codes(7,2,4):
        ....:    print(B); print(B.generator_matrix())
        [4, 1] linear code over GF(2)
        [1 1 1 1]
        [6, 2] linear code over GF(2)
        [1 0 1 0 1 1]
        [0 1 0 1 1 1]

    Generate all self-orthogonal codes of length equal to 8 and
    dimension equal to 4::

        sage: # needs sage.groups
        sage: for B in codes.databases.self_orthogonal_binary_codes(8, 4, equal=True):
        ....:     print(B); print(B.generator_matrix())
        [8, 4] linear code over GF(2)
        [1 0 0 1 0 0 0 0]
        [0 1 0 0 1 0 0 0]
        [0 0 1 0 0 1 0 0]
        [0 0 0 0 0 0 1 1]
        [8, 4] linear code over GF(2)
        [1 0 0 0 1 1 0 1]
        [0 1 0 0 1 0 1 1]
        [0 0 1 0 1 1 1 0]
        [0 0 0 1 0 1 1 1]

    Since all the codes will be self-orthogonal, b must be divisible by
    2::

        sage: list(codes.databases.self_orthogonal_binary_codes(8, 4, 1, equal=True))
        Traceback (most recent call last):
        ...
        ValueError: b (1) must be a positive even integer.
    """
    from sage.rings.finite_rings.finite_field_constructor import FiniteField
    from sage.matrix.constructor import Matrix

    d = int(b)
    if d != b or d % 2 == 1 or d <= 0:
        raise ValueError("b (%s) must be a positive even integer." % b)
    from .linear_code import LinearCode
    from .binary_code import BinaryCode, BinaryCodeClassifier
    if k < 1 or n < 2:
        return
    if equal:
        in_test = lambda M: (M.ncols() - M.nrows()) <= (n-k)
        out_test = lambda C: (C.dimension() == k) and (C.length() == n)
    else:
        in_test = lambda M: True
        out_test = lambda C: True
    if BC is None:
        BC = BinaryCodeClassifier()
    if parent is None:
        for j in range(d, n+1, d):
            M = Matrix(FiniteField(2), [[1]*j])
            if in_test(M):
                for N in self_orthogonal_binary_codes(n, k, d, M, BC, in_test=in_test):
                    if out_test(N):
                        yield N
    else:
        C = LinearCode(parent)
        if out_test(C):
            yield C
        if k == parent.nrows():
            return
        for nn in range(parent.ncols()+1, n+1):
            if in_test(parent):
                for child in BC.generate_children(BinaryCode(parent), nn, d):
                    for N in self_orthogonal_binary_codes(n, k, d, child, BC, in_test=in_test):
                        if out_test(N):
                            yield N



# sage.doctest: needs sage.combinat sage.modules
r"""
Induced crystals

We construct a crystal structure on a set induced by a bijection `\Phi`.

AUTHORS:

- Travis Scrimshaw (2014-05-15): initial implementation
"""

#*****************************************************************************
#       Copyright (C) 2014 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#****************************************************************************

from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.parent import Parent
from sage.structure.element_wrapper import ElementWrapper


class InducedCrystal(UniqueRepresentation, Parent):
    r"""
    A crystal induced from an injection.

    Let `X` be a set and let `C` be crystal and consider any injection
    `\Phi : X \to C`. We induce a crystal structure on `X` by considering
    `\Phi` to be a crystal morphism.

    Alternatively we can induce a crystal structure on some (sub)set of `X`
    by considering an injection `\Phi : C \to X` considered as a crystal
    morphism. This form is also useful when the set `X` is not explicitly
    known.

    INPUT:

    - ``X`` -- the base set
    - ``phi`` -- the map `\Phi`
    - ``inverse`` -- (optional) the inverse map `\Phi^{-1}`
    - ``from_crystal`` -- boolean (default: ``False``); if the induced
      structure is of the second type `\Phi : C \to X`

    EXAMPLES:

    We construct a crystal structure of Gelfand-Tsetlin patterns by going
    through their bijection with semistandard tableaux::

        sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,3))
        sage: G = GelfandTsetlinPatterns(4, 3)
        sage: phi = lambda x: D(x.to_tableau())
        sage: phi_inv = lambda x: G(x.to_tableau())
        sage: I = crystals.Induced(G, phi, phi_inv)
        sage: I.digraph().is_isomorphic(D.digraph(), edge_labels=True)
        True

    Now we construct the above example but inducing the structure going the
    other way (from tableaux to Gelfand-Tsetlin patterns). This can also
    give us more information coming from the crystal. ::

        sage: D2 = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
        sage: G2 = GelfandTsetlinPatterns(4, 1)
        sage: phi2 = lambda x: D2(x.to_tableau())
        sage: phi2_inv = lambda x: G2(x.to_tableau())
        sage: I2 = crystals.Induced(D2, phi2_inv, phi2, from_crystal=True)
        sage: I2.module_generators
        ([[0, 0, 0, 0], [0, 0, 0], [0, 0], [0]],
         [[1, 0, 0, 0], [1, 0, 0], [1, 0], [1]],
         [[1, 1, 0, 0], [1, 1, 0], [1, 1], [1]],
         [[1, 1, 1, 0], [1, 1, 1], [1, 1], [1]],
         [[1, 1, 1, 1], [1, 1, 1], [1, 1], [1]])

    We check an example when the codomain is larger than the domain
    (although here the crystal structure is trivial)::

        sage: P = Permutations(4)
        sage: D = crystals.Tableaux(['A',3], shapes=Partitions(4))
        sage: T = crystals.TensorProduct(D, D)
        sage: phi = lambda p: T(D(RSK(p)[0]), D(RSK(p)[1]))
        sage: phi_inv = lambda d: RSK_inverse(d[0].to_tableau(), d[1].to_tableau(), output='permutation')
        sage: all(phi_inv(phi(p)) == p for p in P) # Check it really is the inverse
        True
        sage: I = crystals.Induced(P, phi, phi_inv)
        sage: I.digraph()
        Digraph on 24 vertices

    We construct an example without a specified inverse map::

        sage: X = Words(2,4)
        sage: L = crystals.Letters(['A',1])
        sage: T = crystals.TensorProduct(*[L]*4)
        sage: Phi = lambda x : T(*[L(i) for i in x])
        sage: I = crystals.Induced(X, Phi)
        sage: I.digraph()
        Digraph on 16 vertices
    """
    @staticmethod
    def __classcall_private__(cls, X, phi, inverse=None, from_crystal=False):
        """
        Normalize input to ensure a unique representation.

        TESTS::

            sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,3))
            sage: G = GelfandTsetlinPatterns(4, 3)
            sage: phi = lambda x: D(x.to_tableau())
            sage: phi_inv = lambda x: G(x.to_tableau())
            sage: I1 = crystals.Induced(G, phi, phi_inv)
            sage: I2 = crystals.Induced(G, phi, phi_inv)
            sage: I1 is I2
            True
        """
        if from_crystal:
            return InducedFromCrystal(X, phi, inverse)

        return super().__classcall__(cls, X, phi, inverse)

    def __init__(self, X, phi, inverse):
        """
        Initialize ``self``.

        TESTS:

        Note that pickling only works when the input functions
        can be pickled::

            sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
            sage: G = GelfandTsetlinPatterns(4, 1)
            sage: def phi(x): return D(x.to_tableau())
            sage: def phi_inv(x): return G(x.to_tableau())
            sage: import __main__
            sage: __main__.phi = phi
            sage: __main__.phi_inv = phi_inv
            sage: I = crystals.Induced(G, phi, phi_inv)
            sage: TestSuite(I).run()
        """
        try:
            codomain = phi.codomain()
        except AttributeError:
            codomain = phi(X.an_element()).parent()

        self._set = X
        self._phi = phi

        if inverse is None:
            try:
                inverse = ~self._phi
            except (TypeError, ValueError):
                try:
                    inverse = self._phi.section()
                except AttributeError:
                    if X.cardinality() == float('inf'):
                        raise ValueError("the inverse map must be defined for infinite sets")
                    self._preimage = {}
                    for x in X:
                        y = phi(x)
                        if y in self._preimage:
                            raise ValueError("the map is not injective")
                        self._preimage[y] = x
                    inverse = self._preimage.__getitem__
        self._inverse = inverse

        self._cartan_type = codomain.cartan_type()
        Parent.__init__(self, category=codomain.category())

        self.module_generators = self

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
            sage: G = GelfandTsetlinPatterns(4, 1)
            sage: def phi(x): return D(x.to_tableau())
            sage: def phi_inv(x): return G(x.to_tableau())
            sage: crystals.Induced(G, phi, phi_inv)
            Crystal of Gelfand-Tsetlin patterns of width 4 and max value 1
             induced by <function phi at 0x...>
        """
        return "Crystal of {} induced by {}".format(self._set, self._phi)

    def _element_constructor_(self, x):
        """
        Construct an element of ``self``.

        EXAMPLES::

            sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
            sage: G = GelfandTsetlinPatterns(4, 1)
            sage: def phi(x): return D(x.to_tableau())
            sage: def phi_inv(x): return G(x.to_tableau())
            sage: I = crystals.Induced(G, phi, phi_inv)
            sage: I([[1,1,0,0],[1,0,0],[1,0],[1]])
            [[1, 1, 0, 0], [1, 0, 0], [1, 0], [1]]
            sage: I(D(3,2,1))
            [[1, 1, 1, 0], [1, 1, 1], [1, 1], [1]]
            sage: I([[1,1,0,0],[1,0,0],[0,1],[1]])
            Traceback (most recent call last):
            ...
            TypeError: unable to convert [[1, 1, 0, 0], [1, 0, 0], [0, 1], [1]] to Crystal of Gelfand-Tsetlin patterns of width 4 and max value 1 induced by <function phi at 0x...>
        """
        if x in self._set:
            return self.element_class(self, self._set(x))

        try:
            return self.element_class(self, self._inverse(x))
        except (TypeError, ValueError, AttributeError):
            raise TypeError("unable to convert {!r} to {}".format(x, self))

    def __contains__(self, x):
        """
        Check if ``x`` is in ``self``.

        EXAMPLES::

            sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
            sage: G = GelfandTsetlinPatterns(4, 1)
            sage: def phi(x): return D(x.to_tableau())
            sage: def phi_inv(x): return G(x.to_tableau())
            sage: I = crystals.Induced(G, phi, phi_inv)
            sage: all(g in I for g in G)
            True
            sage: [[1,1,0,0],[1,0,0],[1,0],[1]] in I
            True
            sage: [[1,1,0,0],[1,0,0],[0,1],[1]] in I
            False
        """
        if isinstance(x, InducedCrystal.Element):
            return x.parent() == self

        return x in self._set

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
            sage: G = GelfandTsetlinPatterns(4, 1)
            sage: def phi(x): return D(x.to_tableau())
            sage: def phi_inv(x): return G(x.to_tableau())
            sage: I = crystals.Induced(G, phi, phi_inv)
            sage: sorted([x for x in I])
            [[[0, 0, 0, 0], [0, 0, 0], [0, 0], [0]],
             [[1, 0, 0, 0], [0, 0, 0], [0, 0], [0]],
             [[1, 0, 0, 0], [1, 0, 0], [0, 0], [0]],
             [[1, 0, 0, 0], [1, 0, 0], [1, 0], [0]],
             [[1, 0, 0, 0], [1, 0, 0], [1, 0], [1]],
             [[1, 1, 0, 0], [1, 0, 0], [0, 0], [0]],
             [[1, 1, 0, 0], [1, 0, 0], [1, 0], [0]],
             [[1, 1, 0, 0], [1, 0, 0], [1, 0], [1]],
             [[1, 1, 0, 0], [1, 1, 0], [1, 0], [0]],
             [[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]],
             [[1, 1, 0, 0], [1, 1, 0], [1, 1], [1]],
             [[1, 1, 1, 0], [1, 1, 0], [1, 0], [0]],
             [[1, 1, 1, 0], [1, 1, 0], [1, 0], [1]],
             [[1, 1, 1, 0], [1, 1, 0], [1, 1], [1]],
             [[1, 1, 1, 0], [1, 1, 1], [1, 1], [1]],
             [[1, 1, 1, 1], [1, 1, 1], [1, 1], [1]]]
        """
        for x in self._set:
            yield self.element_class(self, x)

    def cardinality(self):
        """
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: P = Permutations(4)
            sage: D = crystals.Tableaux(['A',3], shapes=Partitions(4))
            sage: T = crystals.TensorProduct(D, D)
            sage: phi = lambda p: T(D(RSK(p)[0]), D(RSK(p)[1]))
            sage: phi_inv = lambda d: RSK_inverse(d[0].to_tableau(), d[1].to_tableau(), output='permutation')
            sage: I = crystals.Induced(P, phi, phi_inv)
            sage: I.cardinality() == factorial(4)
            True
        """
        return self._set.cardinality()

    class Element(ElementWrapper):
        """
        An element of an induced crystal.
        """

        def e(self, i):
            """
            Return `e_i` of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,3))
                sage: G = GelfandTsetlinPatterns(4, 3)
                sage: phi = lambda x: D(x.to_tableau())
                sage: phi_inv = lambda x: G(x.to_tableau())
                sage: I = crystals.Induced(G, phi, phi_inv)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: elt.e(1)
                sage: elt.e(2)
                [[1, 1, 0, 0], [1, 1, 0], [1, 1], [1]]
                sage: elt.e(3)
            """
            P = self.parent()
            ret = P._phi(self.value).e(i)
            if ret is None:
                return None
            try:
                return self.__class__(P, P._inverse(ret))
            except (ValueError, TypeError, AttributeError):
                return None

        def f(self, i):
            """
            Return `f_i` of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,3))
                sage: G = GelfandTsetlinPatterns(4, 3)
                sage: phi = lambda x: D(x.to_tableau())
                sage: phi_inv = lambda x: G(x.to_tableau())
                sage: I = crystals.Induced(G, phi, phi_inv)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: elt.f(1)
                [[1, 1, 0, 0], [1, 1, 0], [1, 0], [0]]
                sage: elt.f(2)
                sage: elt.f(3)
                [[1, 1, 0, 0], [1, 0, 0], [1, 0], [1]]
            """
            P = self.parent()
            ret = P._phi(self.value).f(i)
            if ret is None:
                return None
            try:
                return self.__class__(P, P._inverse(ret))
            except (ValueError, TypeError, AttributeError):
                return None

        def epsilon(self, i):
            r"""
            Return `\varepsilon_i` of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,3))
                sage: G = GelfandTsetlinPatterns(4, 3)
                sage: phi = lambda x: D(x.to_tableau())
                sage: phi_inv = lambda x: G(x.to_tableau())
                sage: I = crystals.Induced(G, phi, phi_inv)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: [elt.epsilon(i) for i in I.index_set()]
                [0, 1, 0]
            """
            return self.parent()._phi(self.value).epsilon(i)

        def phi(self, i):
            r"""
            Return `\varphi_i` of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,3))
                sage: G = GelfandTsetlinPatterns(4, 3)
                sage: phi = lambda x: D(x.to_tableau())
                sage: phi_inv = lambda x: G(x.to_tableau())
                sage: I = crystals.Induced(G, phi, phi_inv)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: [elt.phi(i) for i in I.index_set()]
                [1, 0, 1]
            """
            return self.parent()._phi(self.value).phi(i)

        def weight(self):
            """
            Return the weight of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,3))
                sage: G = GelfandTsetlinPatterns(4, 3)
                sage: phi = lambda x: D(x.to_tableau())
                sage: phi_inv = lambda x: G(x.to_tableau())
                sage: I = crystals.Induced(G, phi, phi_inv)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: elt.weight()
                (1, 0, 1, 0)
            """
            return self.parent()._phi(self.value).weight()


class InducedFromCrystal(UniqueRepresentation, Parent):
    r"""
    A crystal induced from an injection.

    Alternatively we can induce a crystal structure on some (sub)set of `X`
    by considering an injection `\Phi : C \to X` considered as a crystal
    morphism.

    .. SEEALSO::

        :class:`InducedCrystal`

    INPUT:

    - ``X`` -- the base set
    - ``phi`` -- the map `\Phi`
    - ``inverse`` -- (optional) the inverse map `\Phi^{-1}`

    EXAMPLES:

    We construct a crystal structure on generalized permutations with a
    fixed first row by using RSK::

        sage: C = crystals.Tableaux(['A',3], shape=[2,1])
        sage: def psi(x):
        ....:     ret = RSK_inverse(x.to_tableau(), Tableau([[1,1],[2]]))
        ....:     return (tuple(ret[0]), tuple(ret[1]))
        sage: psi_inv = lambda x: C(RSK(*x)[0])
        sage: I = crystals.Induced(C, psi, psi_inv, from_crystal=True)
    """

    def __init__(self, X, phi, inverse):
        """
        Initialize ``self``.

        TESTS:

        Note that pickling only works when the input functions
        can be pickled::

            sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
            sage: G = GelfandTsetlinPatterns(4, 1)
            sage: def phi(x): return D(x.to_tableau())
            sage: def phi_inv(x): return G(x.to_tableau())
            sage: import __main__
            sage: __main__.phi = phi
            sage: __main__.phi_inv = phi_inv
            sage: I = crystals.Induced(D, phi_inv, phi, from_crystal=True)
            sage: TestSuite(I).run()
        """
        self._crystal = X
        self._phi = phi

        if inverse is None:
            try:
                inverse = ~self._phi
            except (TypeError, ValueError):
                try:
                    inverse = self._phi.section()
                except AttributeError:
                    if X.cardinality() == float('inf'):
                        raise ValueError("the inverse map must be defined for infinite sets")
                    self._preimage = {}
                    for x in X:
                        y = phi(x)
                        if y in self._preimage:
                            raise ValueError("the map is not injective")
                        self._preimage[y] = x
                    inverse = self._preimage.__getitem__
        self._inverse = inverse

        self._cartan_type = X.cartan_type()
        Parent.__init__(self, category=X.category())
        self.module_generators = tuple(self.element_class(self, phi(mg))
                                       for mg in X.module_generators)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
            sage: G = GelfandTsetlinPatterns(4, 1)
            sage: def phi(x): return D(x.to_tableau())
            sage: def phi_inv(x): return G(x.to_tableau())
            sage: crystals.Induced(D, phi_inv, phi, from_crystal=True)
            Crystal induced by <function phi_inv at 0x...> from
             The crystal of tableaux of type ['A', 3] and shape(s)
             [[], [1], [1, 1], [1, 1, 1], [1, 1, 1, 1]]
        """
        return "Crystal induced by {} from {}".format(self._phi, self._crystal)

    def _element_constructor_(self, x):
        """
        Construct an element of ``self``.

        EXAMPLES::

            sage: C = crystals.Tableaux(['A',3], shape=[2,1])
            sage: def psi(x):
            ....:     ret = RSK_inverse(x.to_tableau(), Tableau([[1,1],[2]]))
            ....:     return (tuple(ret[0]), tuple(ret[1]))
            sage: psi_inv = lambda x: C(RSK(*x)[0])
            sage: I = crystals.Induced(C, psi, psi_inv, from_crystal=True)
            sage: I([[1, 1, 2], [2, 2, 1]])
            ((1, 1, 2), (2, 2, 1))
            sage: I(C(2,1,3))
            ((1, 1, 2), (2, 3, 1))
        """
        if x in self._crystal:
            return self.element_class(self, self._phi(self._crystal(x)))

        try:
            return self.element_class(self, self._phi(self._inverse(x)))
        except (TypeError, ValueError, AttributeError):
            raise TypeError("unable to convert {!r} to {}".format(x, self))

    def __contains__(self, x):
        """
        Check if ``x`` is in ``self``.

        EXAMPLES::

            sage: C = crystals.Tableaux(['A',3], shape=[2,1])
            sage: def psi(x):
            ....:     ret = RSK_inverse(x.to_tableau(), Tableau([[1,1],[2]]))
            ....:     return (tuple(ret[0]), tuple(ret[1]))
            sage: psi_inv = lambda x: C(RSK(*x)[0])
            sage: I = crystals.Induced(C, psi, psi_inv, from_crystal=True)
            sage: ((1, 1, 2), (2, 2, 1)) in I
            True
            sage: ((1, 2, 2), (1, 1, 2)) in I
            False
            sage: ((1, 2, 3), (1, 2, 3)) in I
            False
            sage: ((1, 2, 2), (1, 3, 2)) in I
            False
        """
        if isinstance(x, InducedFromCrystal.Element):
            return x.parent() == self

        try:
            y = self._inverse(x)
            return y in self._crystal and self._phi(y) == x
        except (ValueError, TypeError):
            return False

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: C = crystals.Tableaux(['A',2], shape=[2,1])
            sage: def psi(x):
            ....:     ret = RSK_inverse(x.to_tableau(), Tableau([[1,1],[2]]))
            ....:     return (tuple(ret[0]), tuple(ret[1]))
            sage: psi_inv = lambda x: C(RSK(*x)[0])
            sage: I = crystals.Induced(C, psi, psi_inv, from_crystal=True)
            sage: sorted(x for x in I)
            [((1, 1, 2), (1, 2, 1)),
             ((1, 1, 2), (2, 2, 1)),
             ((1, 1, 2), (2, 3, 1)),
             ((1, 1, 2), (3, 3, 1)),
             ((1, 1, 2), (3, 3, 2)),
             ((1, 1, 2), (1, 3, 1)),
             ((1, 1, 2), (1, 3, 2)),
             ((1, 1, 2), (2, 3, 2))]
        """
        for x in self._crystal:
            yield self.element_class(self, self._phi(x))

    def cardinality(self):
        """
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: C = crystals.Tableaux(['A',3], shape=[2,1])
            sage: def psi(x):
            ....:     ret = RSK_inverse(x.to_tableau(), Tableau([[1,1],[2]]))
            ....:     return (tuple(ret[0]), tuple(ret[1]))
            sage: psi_inv = lambda x: C(RSK(*x)[0])
            sage: I = crystals.Induced(C, psi, psi_inv, from_crystal=True)
            sage: I.cardinality() == C.cardinality()
            True
        """
        return self._crystal.cardinality()

    class Element(ElementWrapper):
        """
        An element of an induced crystal.
        """

        def e(self, i):
            """
            Return `e_i` of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
                sage: G = GelfandTsetlinPatterns(4, 1)
                sage: def phi(x): return G(x.to_tableau())
                sage: def phi_inv(x): return D(G(x).to_tableau())
                sage: I = crystals.Induced(D, phi, phi_inv, from_crystal=True)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: elt.e(1)
                sage: elt.e(2)
                [[1, 1, 0, 0], [1, 1, 0], [1, 1], [1]]
                sage: elt.e(3)
            """
            P = self.parent()
            ret = P._inverse(self.value).e(i)
            if ret is None:
                return None
            return self.__class__(P, P._phi(ret))

        def f(self, i):
            """
            Return `f_i` of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
                sage: G = GelfandTsetlinPatterns(4, 1)
                sage: def phi(x): return G(x.to_tableau())
                sage: def phi_inv(x): return D(G(x).to_tableau())
                sage: I = crystals.Induced(D, phi, phi_inv, from_crystal=True)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: elt.f(1)
                [[1, 1, 0, 0], [1, 1, 0], [1, 0], [0]]
                sage: elt.f(2)
                sage: elt.f(3)
                [[1, 1, 0, 0], [1, 0, 0], [1, 0], [1]]
            """
            P = self.parent()
            ret = P._inverse(self.value).f(i)
            if ret is None:
                return None
            return self.__class__(P, P._phi(ret))

        def epsilon(self, i):
            r"""
            Return `\varepsilon_i` of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
                sage: G = GelfandTsetlinPatterns(4, 1)
                sage: def phi(x): return G(x.to_tableau())
                sage: def phi_inv(x): return D(G(x).to_tableau())
                sage: I = crystals.Induced(D, phi, phi_inv, from_crystal=True)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: [elt.epsilon(i) for i in I.index_set()]
                [0, 1, 0]
            """
            return self.parent()._inverse(self.value).epsilon(i)

        def phi(self, i):
            r"""
            Return `\varphi_i` of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
                sage: G = GelfandTsetlinPatterns(4, 1)
                sage: def phi(x): return G(x.to_tableau())
                sage: def phi_inv(x): return D(G(x).to_tableau())
                sage: I = crystals.Induced(D, phi, phi_inv, from_crystal=True)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: [elt.epsilon(i) for i in I.index_set()]
                [0, 1, 0]
            """
            return self.parent()._inverse(self.value).phi(i)

        def weight(self):
            """
            Return the weight of ``self``.

            EXAMPLES::

                sage: D = crystals.Tableaux(['A',3], shapes=PartitionsInBox(4,1))
                sage: G = GelfandTsetlinPatterns(4, 1)
                sage: def phi(x): return G(x.to_tableau())
                sage: def phi_inv(x): return D(G(x).to_tableau())
                sage: I = crystals.Induced(D, phi, phi_inv, from_crystal=True)
                sage: elt = I([[1, 1, 0, 0], [1, 1, 0], [1, 0], [1]])
                sage: elt.weight()
                (1, 0, 1, 0)
            """
            return self.parent()._inverse(self.value).weight()



# sage.doctest: needs sage.combinat
r"""
Finitely presented graded modules

Let `R` be a connected graded algebra.  A finitely presented module
over `R` is a module isomorphic to the cokernel of an `R`-linear
homomorphism `f: F_1 \to F_0` of finitely generated free modules: the
generators of `F_0` correspond to the generators of the module, and
the generators of `F_1` correspond to its relations, via the map `f`.

This module class takes as input a set of generators and relations
and uses them to construct a presentation, using the class
:class:`~sage.modules.fp_graded.free_morphism.FreeGradedModuleMorphism`.
It also allows such a morphism as input.

This package was designed with homological algebra in mind, and its API
focuses on maps rather than objects.  A good example of this is the kernel
function :meth:`sage.modules.fp_graded.morphism.FPModuleMorphism.kernel_inclusion`,
which computes the kernel of a homomorphism `f: M\to N`.  Its return value
is not an instance of the module class, but rather an injective homomorphism
`i: K\to M` with the property that `\operatorname{im}(i) = \ker(f)`.

.. NOTE::

    Some methods here require in addition that `R` be an algebra over a
    field or a PID and that Sage has a description of a basis for `R`.

AUTHORS:

- Robert R. Bruner, Michael J. Catanzaro (2012): Initial version.
- Sverre Lunoee--Nielsen and Koen van Woerden (2019-11-29): Updated the
  original software to Sage version 8.9.
- Sverre Lunoee--Nielsen (2020-07-01): Refactored the code and added
  new documentation and tests.
"""

# *****************************************************************************
#       Copyright (C) 2011 Robert R. Bruner <rrb@math.wayne.edu> and
#                          Michael J. Catanzaro <mike@math.wayne.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.homset import Hom
from sage.categories.morphism import Morphism
from sage.misc.cachefunc import cached_method
from sage.rings.infinity import infinity
from sage.categories.graded_modules import GradedModules
from sage.modules.module import Module
from sage.structure.indexed_generators import IndexedGenerators
from sage.structure.unique_representation import UniqueRepresentation
from sage.misc.lazy_attribute import lazy_attribute
from sage.structure.element import parent

from sage.modules.fp_graded.free_module import FreeGradedModule
from sage.modules.fp_graded.free_element import FreeGradedModuleElement
from sage.modules.fp_graded.element import FPElement

# Note that some of the methods below assume that the base ring is a
# field and that the graded algebra has a chosen free module basis in
# each degree. See for example :meth:`basis_elements`,
# :meth:`element_from_coordinates`, :meth:`__getitem__`,
# :meth:`vector_presentation`, and possibly others.


class FPModule(UniqueRepresentation, IndexedGenerators, Module):
    r"""
    A finitely presented module over a connected graded algebra.

    INPUT:

    One of the following:

    - ``arg0`` -- a morphism such that the module is the cokernel, or
      a free graded module, in which case the output is the same
      module, viewed as finitely presented

    Otherwise:

    - ``arg0`` -- the graded connected algebra over which the module is
      defined; this algebra must be equipped with a graded basis

    - ``generator_degrees`` -- tuple of integer degrees

    - ``relations`` -- tuple of relations; a relation is a tuple of
      coefficients `(c_1, \ldots, c_n)`, ordered so that they
      correspond to the module generators, that is, such a tuple
      corresponds to the relation

      .. MATH::

          c_1 g_1 + \ldots + c_n g_n = 0

      if the generators are `(g_1, \ldots, g_n)`

    EXAMPLES::

        sage: from sage.modules.fp_graded.module import FPModule

        sage: E.<x,y> = ExteriorAlgebra(QQ)
        sage: M = FPModule(E, [0, 1], [[x, 1]])
        sage: a, b = M.generators()
        sage: x*a + b == 0
        True
        sage: (x*a + b).normalize()
        0

        sage: A3 = SteenrodAlgebra(2, profile=(4,3,2,1))
        sage: M = FPModule(A3, [0, 1], [[Sq(2), Sq(1)]])
        sage: M.generators()
        (g[0], g[1])
        sage: M.relations()
        (Sq(2)*g[0] + Sq(1)*g[1],)
        sage: M.is_trivial()
        False

        sage: Z = FPModule(A3, [])
        sage: Z.generators()
        ()
        sage: Z.relations()
        ()
        sage: Z.is_trivial()
        True

        sage: from sage.modules.fp_graded.free_module import FreeGradedModule
        sage: F = FreeGradedModule(E, [0, 1])
        sage: one = Hom(F, F).identity()
        sage: Z = FPModule(one)
        sage: Z.is_trivial()
        True

        sage: FPModule(E.free_graded_module([0, 1]))
        Free graded left module on 2 generators over The exterior algebra of rank 2 over Rational Field
    """
    @staticmethod
    def __classcall__(cls, arg0, generator_degrees=None, relations=(), names=None):
        r"""
        Normalize input to ensure a unique representation.

        TESTS::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A3 = SteenrodAlgebra(2, profile=(4,3,2,1))
            sage: M1.<m0,m1> = FPModule(A3, [0, 1], [[Sq(2), Sq(1)]])
            sage: M2 = FPModule(A3, (0, 1), [[Sq(2), Sq(1)]], names='m0,m1')
            sage: M1 is M2
            True
        """
        if names is not None:
            from sage.structure.category_object import normalize_names
            names = normalize_names(-1, names)

        # If given a morphism, then that defines a module
        if isinstance(arg0, Morphism):
            if arg0.is_zero():
                # This morphism defines a free module, so just use the codomain
                return arg0.codomain()
            return super().__classcall__(cls, arg0, names=names)

        if isinstance(arg0, FreeGradedModule):
            return arg0

        if generator_degrees is None:
            raise ValueError("the generator_degrees must be specified")

        # The free module on the generators of the module.
        generator_module = arg0.free_graded_module(generator_degrees, names=names)

        if not relations:  # If we are trivially free, then just use the free module
            return generator_module

        # Use the coefficients given for the relations and make module elements
        # from them.  Filter out the zero elements, as they are redundant.
        rels = [v for r in relations
                if not (v := generator_module(r)).is_zero()]

        # The free module for the relations of the module.
        relations_module = arg0.free_graded_module(tuple([r.degree()
                                                          for r in rels]))

        # The module we want to model is the cokernel of the following morphism
        j = Hom(relations_module, generator_module)(rels)
        return super().__classcall__(cls, j, names=names)

    def __init__(self, j, names):
        r"""
        Create a finitely presented module over a connected graded algebra.

        TESTS::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A3 = SteenrodAlgebra(2, profile=(4,3,2,1))
            sage: M = FPModule(A3, [0, 1], [[Sq(2), Sq(1)]])
            sage: TestSuite(M).run()

        Checking containment::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: M = FPModule(SteenrodAlgebra(2), [0,1], [[Sq(4), Sq(3)]])
            sage: x = M([Sq(1), 1])
            sage: x in M
            True
            sage: N = FPModule(SteenrodAlgebra(2), [0], [[Sq(2)]])
            sage: y = Sq(2) * N.generator(0)
            sage: y in M
            False
        """
        self._j = j
        codomain = j.codomain()

        self._generator_degrees = codomain.generator_degrees()
        algebra = codomain.base_ring()

        # Call the base class constructors.
        keys = j.codomain().basis().keys()
        cat = GradedModules(algebra).FinitelyPresented()
        IndexedGenerators.__init__(self, keys)
        Module.__init__(self, algebra, category=cat, names=names)

        from sage.combinat.family import Family
        self._spanning_set = Family(self._indices, self.monomial)

    Element = FPElement

    def defining_homomorphism(self):
        """
        Return the homomorphism defining ``self``.

        ``self`` is a finitely presented module defined as the
        cokernel of a map `j: F_0 \to F_1` of free modules, and this
        returns that map.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: E.<x,y> = ExteriorAlgebra(QQ)
            sage: M = FPModule(E, [0, 1], [[x, 1]])
            sage: M.defining_homomorphism()
            Module morphism:
              From: Free graded left module on 1 generator over The exterior algebra of rank 2 over Rational Field
              To:   Free graded left module on 2 generators over The exterior algebra of rank 2 over Rational Field
              Defn: g[1] |--> x*g[0] + g[1]
        """
        return self._j

    def _free_module(self):
        """
        Return the free module of which ``self`` is a quotient.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra()
            sage: M.<x,y> = FPModule(A, [0, 1], [[Sq(2), Sq(1)]])
            sage: M.generators()
            (x, y)
            sage: F = M._free_module()
            sage: F.generators()
            (x, y)
        """
        return self._j.codomain()

    def change_ring(self, algebra):
        r"""
        Change the base ring of ``self``.

        INPUT:

        - ``algebra`` -- a connected graded algebra

        OUTPUT:

        The finitely presented module over ``algebra`` defined with the
        exact same number of generators of the same degrees and relations
        as ``self``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))

            sage: M = FPModule(A, [0,1], [[Sq(2), Sq(1)]])
            sage: N = M.change_ring(A2); N
            Finitely presented left module on 2 generators and 1 relation over
             sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]

        Changing back yields the original module::

            sage: N.change_ring(A) is M
            True

        TESTS:

        Subclasses preserve their type::

            sage: from sage.modules.fp_graded.steenrod.module import SteenrodFPModule
            sage: A = SteenrodAlgebra(2)
            sage: A1 = SteenrodAlgebra(2, profile=(2,1))

            sage: M = SteenrodFPModule(A, [0,1], [[Sq(2), Sq(1)]])
            sage: N = M.change_ring(A1)
            sage: isinstance(N, SteenrodFPModule)
            True

        Changing back yields the original module::

            sage: N.change_ring(A) is M
            True
        """
        return type(self).__base__(self._j.change_ring(algebra), names=self._names)

    def _from_dict(self, d, coerce=False, remove_zeros=True):
        r"""
        Construct an element of ``self`` from an ``{index: coefficient}``
        dictionary.

        INPUT:

        - ``d`` -- dictionary

        This code is taken from the method of the same name for
        ``sage.combinat.free_module.FreeModule``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: M = FPModule(SteenrodAlgebra(2), [0,1], [[Sq(4), Sq(3)]])
            sage: x = M.an_element()
            sage: M._from_dict(x.monomial_coefficients()) == x
            True
        """
        assert isinstance(d, dict)
        if coerce:
            R = self.base_ring()
            d = {key: R(coeff) for key, coeff in d.items()}
        if remove_zeros:
            d = {key: coeff for key, coeff in d.items() if coeff}
        return self.element_class(self, d)

    def _monomial(self, index):
        """
        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: M = FPModule(SteenrodAlgebra(2), [0,1], [[Sq(4), Sq(3)]])
            sage: M._monomial(0)
            g[0]

        TESTS::

            sage: M.monomial(3)
            Traceback (most recent call last):
             ...
            ValueError: index not in indices
        """
        if index not in self._indices:
            raise ValueError("index not in indices")
        return self._from_dict({index: self.base_ring().one()}, remove_zeros=False)

    @lazy_attribute
    def monomial(self):
        """
        Return the basis element indexed by ``i``.

        INPUT:

        - ``i`` -- an element of the index set

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: M = FPModule(SteenrodAlgebra(2), [0,1], [[Sq(4), Sq(3)]])
            sage: M.monomial(0)
            g[0]
            sage: M.monomial(1)
            g[1]
        """
        # Should use a real Map, as soon as combinatorial_classes are enumerated sets, and therefore parents
        from sage.categories.poor_man_map import PoorManMap
        return PoorManMap(self._monomial, domain=self._indices, codomain=self, name="Term map")

    @cached_method
    def zero(self):
        r"""
        Return the zero element.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: M = FPModule(SteenrodAlgebra(2), [0,1], [[Sq(4), Sq(3)]])
            sage: M.zero()
            0
        """
        return self.element_class(self, {})

    def _element_constructor_(self, x):
        r"""
        Construct any element of ``self``.

        INPUT:

        - ``x`` -- tuple of coefficients, an element of FPModule, or the
          zero integer constant

        OUTPUT:

        An instance of the element class with coefficients from ``x``, the
        element ``x`` if it already was an element, or the zero element.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: M.<a0,a2,a4> = FPModule(A, [0,2,4], [[Sq(4), Sq(2), 0]])

        Creating an element from coefficients::

            sage: e = M((Sq(6), 0, Sq(2))); e
            Sq(6)*a0 + Sq(2)*a4
            sage: e in M
            True

        Creating the zero element::

            sage: z = M(0); z
            0
            sage: z.is_zero()
            True

        Idempotent on elements belonging to the module::

            sage: M(e)
            Sq(6)*a0 + Sq(2)*a4
            sage: e is M(e)
            True
        """
        if parent(x) is self:
            return x
        if not self._generator_degrees:  # the trivial module
            return self.zero()
        if not x:
            return self.zero()
        B = self._spanning_set
        if isinstance(x, FreeGradedModuleElement):
            if x.parent() is self._free_module():
                # x.parent() should have the same generator list as self.
                coeffs = x.monomial_coefficients()
                return sum(coeffs[idx] * B[idx] for idx in coeffs)
            raise ValueError("element is not in this module")
        return self._from_dict({b: c for (c, b) in zip(x, self._indices) if c},
                               remove_zeros=False)

    def _repr_(self):
        r"""
        Construct a string representation of ``self``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FPModule(A, [0,2,4], [[Sq(4),Sq(2),0]]); M
            Finitely presented left module on 3 generators and 1 relation over
             mod 2 Steenrod algebra, milnor basis
            sage: N = FPModule(A, [0,1], [[Sq(2),Sq(1)], [Sq(2)*Sq(1),Sq(2)]]); N
            Finitely presented left module on 2 generators and 2 relations over
             mod 2 Steenrod algebra, milnor basis
            sage: F = FPModule(A, [2]); F
            Free graded left module on 1 generator over
             mod 2 Steenrod algebra, milnor basis
        """
        return "Finitely presented left module on %s generator%s and %s relation%s over %s"\
            % (len(self._free_module().generator_degrees()),
               "" if len(self._free_module().generator_degrees()) == 1 else "s",
               len(self._j.values()), "" if len(self._j.values()) == 1 else "s",
               self.base_ring())

    def _repr_term(self, m):
        """
        Return a string representing the generator indexed by ``m``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FPModule(A, [0,2,4], [[Sq(4),Sq(2),0]])
            sage: M._repr_term(4)
            'g[4]'
        """
        return self._free_module()._repr_term(m)

    def _latex_term(self, m):
        """
        Return a LaTeX representing the generator indexed by ``m``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FPModule(A, [0,2,4], [[Sq(4),Sq(2),0]])
            sage: M._latex_term(4)
            'g_{4}'
        """
        return self._free_module()._latex_term(m)

    def connectivity(self):
        r"""
        The connectivity of ``self``.

        Since a finitely presented module over a connected algebra is in
        particular bounded below, the connectivity is an integer when the
        module is non-trivial, and `+\infty` when the module is trivial.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)

            sage: M = FPModule(A, [0,2,4], [[0, Sq(5), Sq(3)], [Sq(7), 0, Sq(2)*Sq(1)]])
            sage: M.connectivity()
            0

            sage: G = FPModule(A, [0,2], [[1,0]])
            sage: G.connectivity()
            2

        TESTS::

            sage: C = FPModule(SteenrodAlgebra(2, profile=(3,2,1)), [0], relations=[[Sq(1)], [0]])
            sage: C.connectivity()
            0

            sage: F = FPModule(A, [-1])
            sage: F.connectivity()
            -1

            sage: F = FPModule(A, [])
            sage: F.connectivity()
            +Infinity

            sage: F = FPModule(A, [0], [[1]])
            sage: F.connectivity()
            +Infinity
        """
        # In case there are no relations, the connectivity is the equal to
        # the connectivity of the free module on the generators.
        if self._j._degree is None:
            return self._free_module().connectivity()

        # We must check that the generator(s) in the free generator module are
        # not hit by relations, since we are not guaranteed that the
        # presentation we have is minimal.
        X = sorted(self.generator_degrees())

        previous = None
        for k in X:
            if previous is not None and k == previous:
                continue
            if not self._j.vector_presentation(k - self._j._degree).is_surjective():
                return k
            previous = k

        return infinity

    def is_trivial(self) -> bool:
        r"""
        Return ``True`` if ``self`` is isomorphic to the trivial module
        and ``False`` otherwise.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))

            sage: M = FPModule(A2, [])
            sage: M.is_trivial()
            True

            sage: N = FPModule(A, [1,2])
            sage: N.is_trivial()
            False

            sage: P = FPModule(A, [1,2], [[1,0], [0,1]])
            sage: P.is_trivial()
            True

        TESTS::

            sage: C = FPModule(SteenrodAlgebra(2, profile=(3,2,1)), [0], [[Sq(1)], [0]])
            sage: C.is_trivial()
            False

            sage: C = FPModule(SteenrodAlgebra(2), [0], [[Sq(1)], [1]])
            sage: C.is_trivial()
            True
        """
        return self.connectivity() == infinity

    def has_relations(self) -> bool:
        r"""
        Return ``True`` if no relations are defined, and ``False``
        otherwise.

        .. NOTE::

            This module is free if this function returns ``False``, but a free
            module can have (redundant) relations.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))

            sage: F = FPModule(A2, [1,2])
            sage: F.has_relations()
            False

            sage: M = FPModule(A2, [1,2], [[Sq(2), Sq(1)]])
            sage: M.has_relations()
            True

        A free module constructed with a redundant generator and relation::

            sage: N = FPModule(A2, [0, 0], [[0, 1]])
            sage: N.has_relations()
            True
            sage: # Computing a minimal presentation reveals an
            ....: # isomorphic module with no relations.
            sage: N_min = N.minimal_presentation().domain()
            sage: N_min.has_relations()
            False
        """
        return not self._j.is_zero()

    def an_element(self, n=None):
        r"""
        An element of this module.

        This function chooses deterministically an element, i.e. the output
        depends only on the module and its input ``n``.

        INPUT:

        - ``n`` -- (optional) the degree of the element to construct

        OUTPUT: a module element of the given degree

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: M = FPModule(A2, [0,2,4], [[0, Sq(5), Sq(3)], [Sq(7), 0, Sq(2)*Sq(1)]])

            sage: [M.an_element(i) for i in range(10)]
            [g[0],
             Sq(1)*g[0],
             Sq(2)*g[0] + g[2],
             Sq(0,1)*g[0] + Sq(1)*g[2],
             Sq(1,1)*g[0] + Sq(2)*g[2] + g[4],
             Sq(2,1)*g[0] + Sq(0,1)*g[2] + Sq(1)*g[4],
             Sq(0,2)*g[0] + Sq(1,1)*g[2] + Sq(2)*g[4],
             Sq(0,0,1)*g[0] + Sq(2,1)*g[2] + Sq(0,1)*g[4],
             Sq(1,0,1)*g[0] + Sq(6)*g[2] + Sq(1,1)*g[4],
             Sq(2,0,1)*g[0] + Sq(4,1)*g[2] + Sq(2,1)*g[4]]
        """
        a_free_element = self._free_module().an_element(n)
        return self(a_free_element)

    @cached_method
    def basis_elements(self, n, verbose=False):
        r"""
        Return a basis for the free module of degree ``n`` module elements.

        .. NOTE::

            Suppose ``self`` is a module over the graded algebra `A` with
            base ring `R`. This returns a basis as a free module over `R`.

        INPUT:

        - ``n`` -- integer
        - ``verbose`` -- boolean (default: ``False``); controls whether log
          messages should be emitted

        OUTPUT:

        A list of homogeneous module elements of degree ``n`` which is
        a basis for the free module of all degree ``n`` module elements.

        .. SEEALSO::

            :meth:`vector_presentation`, :meth:`element_from_coordinates`

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: M.<m0,m2> = FPModule(A2, [0,2], [[Sq(4), Sq(2)], [0, Sq(6)]])

            sage: M.basis_elements(4)
            (Sq(1,1)*m0, Sq(4)*m0)

            sage: M.basis_elements(5)
            (Sq(2,1)*m0, Sq(5)*m0, Sq(0,1)*m2)

            sage: M.basis_elements(25)
            ()

            sage: M.basis_elements(0)
            (m0,)

            sage: M.basis_elements(2)
            (Sq(2)*m0, m2)

        TESTS::

            sage: Z0 = FPModule(A2, [])
            sage: Z0.basis_elements(n=10)
            ()

            sage: Z1 = FPModule(A2, [1], [[1]])
            sage: Z1.basis_elements(n=10)
            ()
        """
        return tuple([self.element_from_coordinates(x, n) for
                      x in self.vector_presentation(n, verbose).basis()])

    @cached_method
    def element_from_coordinates(self, coordinates, n):
        r"""
        Return the module element in degree ``n`` having the given coordinates
        with respect to the basis returned by :meth:`basis_elements`.

        This function is inverse to
        :meth:`sage.modules.fp_graded.element.FPElement.vector_presentation`.

        INPUT:

        - ``coordinates`` -- a vector of coordinates
        - ``n`` -- the degree of the element to construct

        OUTPUT:

        A module element of degree ``n`` having the given coordinates
        with respect to the basis returned by :meth:`basis_elements`.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FPModule(A, [0], [[Sq(4)], [Sq(7)], [Sq(4)*Sq(9)]])

            sage: M.vector_presentation(12).dimension()
            3
            sage: x = M.element_from_coordinates((0,1,0), 12); x
            Sq(0,4)*g[0]

        Applying the inverse function brings us back to the coordinate form::

            sage: x.vector_presentation()
            (0, 1, 0)

        TESTS::

            sage: M.element_from_coordinates((0,1,0,0), 12)
            Traceback (most recent call last):
             ...
            ValueError: the given coordinate vector has incorrect length (4);
             it should have length 3

        .. SEEALSO::

            :meth:`sage.modules.fp_graded.module.FPModule.vector_presentation`
        """
        M_n = self.vector_presentation(n)

        if len(coordinates) != M_n.dimension():
            raise ValueError('the given coordinate vector has incorrect length (%d); '
                             'it should have length %d' % (len(coordinates), M_n.dimension()))

        free_element = self._free_module().element_from_coordinates(
            M_n.lift(coordinates), n)

        return self(free_element.dense_coefficient_list())

    @cached_method
    def vector_presentation(self, n, verbose=False):
        r"""
        Return a free module isomorphic to the free module of module
        elements of degree ``n``.

        INPUT:

        - ``n`` -- the degree of the presentation

        OUTPUT: a vector space

        .. SEEALSO::

            :meth:`basis_elements`, :meth:`element_from_coordinates`

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FPModule(A, [0,2,4], [[Sq(4),Sq(2),0]])

            sage: V = M.vector_presentation(4)
            sage: V.dimension()
            3

            sage: len(M.basis_elements(4))
            3

        TESTS::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FPModule(A, [0,2,4], [[Sq(4),Sq(2),0]])
            sage: M[4].dimension()
            3
        """
        # Get the vector space presentation of the free module on the
        # module generators.
        F_n = self._free_module().vector_presentation(n)

        # Compute the sub vector space generated by the relations.
        spanning_set = []

        if verbose:
            num_total_iterations = 0
            for relation in self._j.values():
                if relation.is_zero():
                    continue

                num_total_iterations += len(self.base_ring().basis(n - relation.degree()))

            progress = 0
            iteration_count = 0

        for relation in self._j.values():
            if relation.is_zero():
                continue

            for a in self.base_ring().basis(n - relation.degree()):
                if verbose:
                    iteration_count += 1
                    prog = int(100 * iteration_count / num_total_iterations)
                    if prog > progress:
                        progress = prog
                        print('Progress: %d/100' % prog)

                # assert: isinstance(FreeElement, relation)
                v = (a * relation).vector_presentation()
                if v is not None:
                    spanning_set.append(v)

        R_n = F_n.submodule(spanning_set)

        # Return the quotient of the free part by the relations.
        return F_n / R_n

    __getitem__ = vector_presentation

    def _Hom_(self, Y, category):
        r"""
        The internal hook used by the free function
        :meth:`sage.categories.homset.hom.Hom` to create homsets
        involving ``self``.

        TESTS::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: F = FPModule(A, [1,3])
            sage: L = FPModule(A, [2,3], [[Sq(2),Sq(1)], [0,Sq(2)]])

            sage: Hom(F, L)
            Set of Morphisms from Free graded left module on 2 generators ...

            sage: M = A.free_graded_module((2,1))
            sage: Hom(F, M)
            Set of Morphisms from Free graded left module on 2 generators ...
        """
        from .homspace import FPModuleHomspace
        if not isinstance(Y, (FPModule, FreeGradedModule)):
            raise ValueError('cannot create homspace between incompatible types:\n%s  ->\n%s' % (self.__class__, type(Y)))
        if Y.base_ring() != self.base_ring():
            raise ValueError('the modules are not defined over the same base ring')

        return FPModuleHomspace(self, Y, category)

    def generator_degrees(self):
        r"""
        Return the degrees of the generators for ``self``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A4 = SteenrodAlgebra(2, profile=(4,3,2,1))
            sage: N = FPModule(A4, [0, 1], [[Sq(2), Sq(1)]])

            sage: N.generator_degrees()
            (0, 1)
        """
        return self._generator_degrees

    def generators(self):
        r"""
        Return the generators of ``self``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A4 = SteenrodAlgebra(2, profile=(4,3,2,1))

            sage: M = FPModule(A4, [0,0,2,3])
            sage: M.generators()
            (g[0, 0], g[0, 1], g[2, 0], g[3, 0])

            sage: N = FPModule(A4, [0, 1], [[Sq(2), Sq(1)]], names='h')
            sage: N.generators()
            (h[0], h[1])

            sage: Z = FPModule(A4, [])
            sage: Z.generators()
            ()
        """
        return tuple([self.generator(i) for i in range(len(self.generator_degrees()))])

    gens = generators

    def generator(self, index):
        r"""
        Return the module generator with the given index.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A4 = SteenrodAlgebra(2, profile=(4,3,2,1))

            sage: M = FPModule(A4, [0,2,3])
            sage: M.generator(0)
            g[0]

            sage: N = FPModule(A4, [0, 1], [[Sq(2), Sq(1)]], names='h')
            sage: N.generator(1)
            h[1]
        """
        return self(self._free_module().generator(index))

    gen = generator

    def relations(self):
        r"""
        Return the relations of ``self``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A4 = SteenrodAlgebra(2, profile=(4,3,2,1))

            sage: M = FPModule(A4, [0,2,3])
            sage: M.relations()
            ()

            sage: N = FPModule(A4, [0, 1], [[Sq(2), Sq(1)]])
            sage: N.relations()
            (Sq(2)*g[0] + Sq(1)*g[1],)

            sage: Z = FPModule(A4, [])
            sage: Z.relations()
            ()
        """
        return self._j.values()

    def relation(self, index):
        r"""
        Return the module relation of the given index.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A4 = SteenrodAlgebra(2, profile=(4,3,2,1))
            sage: N = FPModule(A4, [0, 1], [[Sq(2), Sq(1)]])
            sage: N.relation(0)
            Sq(2)*g[0] + Sq(1)*g[1]
        """
        return self._j.values()[index]

    def minimal_presentation(self, top_dim=None, verbose=False):
        r"""
        Return a minimal presentation of ``self``.

        OUTPUT:

        An isomorphism `M \to S`, where `M` has minimal presentation
        and `S` is ``self``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))

            sage: M = FPModule(A2, [0,1], [[Sq(2),Sq(1)],[0,Sq(2)],[Sq(3),0]])
            sage: i = M.minimal_presentation()
            sage: M_min = i.domain()

        ``i`` is an isomorphism between ``M_min`` and ``M``::

            sage: i.codomain() is M
            True
            sage: i.is_injective()
            True
            sage: i.is_surjective()
            True

        There are more relations in ``M`` than in ``M_min``::

            sage: M.relations()
            (Sq(2)*g[0] + Sq(1)*g[1], Sq(2)*g[1], Sq(3)*g[0])
            sage: M_min.relations()
            (Sq(2)*g[0] + Sq(1)*g[1], Sq(2)*g[1])

        TESTS::

            sage: T = FPModule(A2, [0], [[1]])
            sage: T_min = T.minimal_presentation().domain()
            sage: T_min.is_trivial()
            True
            sage: T_min
            Free graded left module on 0 generators over ...

        Similar example but using
        :class:`sage.modules.fp_graded.steenrod.module.SteenrodFPModule`::

            sage: from sage.modules.fp_graded.steenrod.module import SteenrodFPModule
            sage: A = SteenrodAlgebra(2)
            sage: M = SteenrodFPModule(A, [0,1], [[Sq(2),Sq(1)],[0,Sq(2)],[Sq(3),0]])

            sage: i = M.minimal_presentation()
            sage: i.codomain() is M
            True

            sage: i.is_injective()
            True
            sage: i.is_surjective()
            True

            sage: i.domain().relations()
            (Sq(2)*g[0] + Sq(1)*g[1], Sq(2)*g[1])

            sage: i.codomain().relations()
            (Sq(2)*g[0] + Sq(1)*g[1], Sq(2)*g[1], Sq(3)*g[0])
        """
        return Hom(self, self).identity().image(top_dim, verbose)

    def suspension(self, t):
        r"""
        Return the suspension of ``self`` by degree ``t``.

        INPUT:

        - ``t`` -- integer degree by which the module is suspended

        OUTPUT:

        A module which is identical to this module by a shift of
        degrees by the integer ``t``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))

            sage: Y = FPModule(A2, [0], [[Sq(1)]])
            sage: X = Y.suspension(4)
            sage: X.generator_degrees()
            (4,)
            sage: X.relations()
            (Sq(1)*g[4],)

            sage: M = FPModule(A, [2,3], [[Sq(2), Sq(1)], [0, Sq(2)]])
            sage: Q = M.suspension(1)
            sage: Q.generator_degrees()
            (3, 4)
            sage: Q.relations()
            (Sq(2)*g[3] + Sq(1)*g[4], Sq(2)*g[4])
            sage: Q = M.suspension(-3)
            sage: Q.generator_degrees()
            (-1, 0)
            sage: Q = M.suspension(0)
            sage: Q.generator_degrees()
            (2, 3)
        """
        relations = tuple([r.dense_coefficient_list() for r in self._j._values])
        return type(self).__base__(self.base_ring(),
                                   tuple([g + t for g in self._generator_degrees]),
                                   relations)

    def submodule_inclusion(self, spanning_elements):
        r"""
        Return the inclusion morphism of the submodule of ``self`` spanned
        by the given elements.

        INPUT:

        - ``spanning_elements`` -- an iterable of elements

        OUTPUT: the inclusion of the submodule into this module

        Because a submodule of a finitely presented module need not be
        finitely presented, this method will only work if the
        underlying algebra is finite-dimensional. Indeed, the current
        implementation only works if the algebra has a ``top_class``
        method, which gets used in
        :meth:`sage.modules.fp_graded.morphism._resolve_kernel`.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))

            sage: M = FPModule(A2, [0,1], [[Sq(2),Sq(1)]])
            sage: i = M.submodule_inclusion([M.generator(0)])
            sage: i.codomain() is M
            True
            sage: i.is_injective()
            True
            sage: i.domain().generator_degrees()
            (0,)
            sage: i.domain().relations()
            (Sq(3)*g[0],)
        """
        # Create the graded module on the set of spanning elements.
        degs = [x.degree() for x in spanning_elements]
        F = type(self).__base__(self.base_ring(), tuple(degs))

        # The submodule is the module generated by the spanning elements.
        return Hom(F, self)(spanning_elements).image()

    def resolution(self, k, top_dim=None, verbose=False):
        r"""
        Return a free resolution of this module of length ``k``.

        INPUT:

        - ``k`` -- nonnegative integer
        - ``top_dim`` -- stop the computation at this degree
          (default: ``None``, but required if the algebra is
          not finite-dimensional)
        - ``verbose`` -- boolean (default: ``False``); control if
          log messages should be emitted

        OUTPUT: list of homomorphisms `[\epsilon, f_1, \ldots, f_k]` such that

        .. MATH::

            f_i: F_i \to F_{i-1} \text{ for } 1 < i \leq k,
            \qquad
            \epsilon: F_0 \to M,

        where each `F_i` is a finitely generated free module, and the
        sequence

        .. MATH::

            F_k \xrightarrow{\mathit{f_k}} F_{k-1}
            \xrightarrow{\mathit{f_{k-1}}} \ldots \rightarrow F_0
            \xrightarrow{\epsilon} M \rightarrow 0

        is exact. Note that the 0th element in this list is the map
        `\epsilon`, and the rest of the maps are between free
        modules.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule

            sage: E.<x,y> = ExteriorAlgebra(QQ)
            sage: M = FPModule(E, [0], [[x], [y]])
            sage: res = M.resolution(3); res
            [Module morphism:
               From: Free graded left module on 1 generator over The exterior algebra of rank 2 over Rational Field
               To:   Finitely presented left module on 1 generator and 2 relations over The exterior algebra of rank 2 over Rational Field
               Defn: g[0] |--> g[0],
             Module morphism:
               From: Free graded left module on 2 generators over The exterior algebra of rank 2 over Rational Field
               To:   Free graded left module on 1 generator over The exterior algebra of rank 2 over Rational Field
               Defn: g[1, 0] |--> x*g[0]
                     g[1, 1] |--> y*g[0],
             Module morphism:
               From: Free graded left module on 3 generators over The exterior algebra of rank 2 over Rational Field
               To:   Free graded left module on 2 generators over The exterior algebra of rank 2 over Rational Field
               Defn: g[2, 0] |--> x*g[1, 0]
                     g[2, 1] |--> y*g[1, 0] + x*g[1, 1]
                     g[2, 2] |--> y*g[1, 1],
             Module morphism:
               From: Free graded left module on 4 generators over The exterior algebra of rank 2 over Rational Field
               To:   Free graded left module on 3 generators over The exterior algebra of rank 2 over Rational Field
               Defn: g[3, 0] |--> x*g[2, 0]
                     g[3, 1] |--> y*g[2, 0] + x*g[2, 1]
                     g[3, 2] |--> y*g[2, 1] + x*g[2, 2]
                     g[3, 3] |--> y*g[2, 2]]
            sage: all((res[i] * res[i+1]).is_zero() for i in range(len(res)-1))
            True

            sage: e = SymmetricFunctions(QQ).e()
            sage: M = FPModule(e, [0], [[e[2]+e[1,1]], [e[1,1]]])
            sage: res = M.resolution(3, top_dim=10)
            sage: all((res[i] * res[i+1]).is_zero() for i in range(2))
            True
            sage: res[-1].domain().is_trivial()
            True
            sage: M = FPModule(e, [0,2], [[e[2]+e[1,1], 0], [e[2,1], e[1]], [0, e[1,1]]])
            sage: res = M.resolution(3, top_dim=10)
            sage: all((res[i] * res[i+1]).is_zero() for i in range(2))
            True
            sage: res[-1].domain().is_trivial()
            True

            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: M = FPModule(A2, [0,1], [[Sq(2), Sq(1)]])
            sage: M.resolution(0)
            [Module morphism:
               From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               To:   Finitely presented left module on 2 generators and 1 relation over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               Defn: g[0] |--> g[0]
                     g[1] |--> g[1]]
            sage: res = M.resolution(4, verbose=True)
            Computing f_1 (1/4)
            Computing f_2 (2/4)
            Computing using the profile:
            (2, 1)
            Resolving the kernel in the range of dimensions [2, 8]: 2 3 4 5 6 7 8.
            Computing f_3 (3/4)
            Computing using the profile:
            (2, 1)
            Resolving the kernel in the range of dimensions [8, 14]: 8 9 10 11 12 13 14.
            Computing f_4 (4/4)
            Computing using the profile:
            (2, 1)
            Resolving the kernel in the range of dimensions [9, 16]: 9 10 11 12 13 14 15 16.
            sage: len(res)
            5
            sage: res
            [Module morphism:
               From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               To:   Finitely presented left module on 2 generators and 1 relation over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               Defn: g[0] |--> g[0]
                     g[1] |--> g[1],
             Module morphism:
               From: Free graded left module on 1 generator over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               To:   Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               Defn: g[2] |--> Sq(2)*g[0] + Sq(1)*g[1],
             Module morphism:
               From: Free graded left module on 1 generator over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               To:   Free graded left module on 1 generator over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               Defn: g[8] |--> Sq(3,1)*g[2],
             Module morphism:
               From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               To:   Free graded left module on 1 generator over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               Defn: g[9] |--> Sq(1)*g[8]
                     g[10] |--> Sq(2)*g[8],
             Module morphism:
               From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               To:   Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
               Defn: g[10] |--> Sq(1)*g[9]
                     g[12] |--> Sq(0,1)*g[9] + Sq(2)*g[10]]
            sage: for i in range(len(res)-1):
            ....:     assert (res[i] * res[i+1]).is_zero(), 'the result is not a complex'

        We construct `\GF{3}` as a `\ZZ`-module (with trivial grading
        concentrated in degree 0) and compute its resolution::

            sage: E = ExteriorAlgebra(ZZ, 0)
            sage: M = FPModule(E, [0], [[3]])
            sage: res = M.resolution(3)
            sage: res
            [Module morphism:
               From: Free graded left module on 1 generator over The exterior algebra of rank 0 over Integer Ring
               To:   Finitely presented left module on 1 generator and 1 relation over The exterior algebra of rank 0 over Integer Ring
               Defn: g[0] |--> g[0],
             Module endomorphism of Free graded left module on 1 generator over The exterior algebra of rank 0 over Integer Ring
               Defn: g[0] |--> 3*g[0],
             Module morphism:
               From: Free graded left module on 0 generators over The exterior algebra of rank 0 over Integer Ring
               To:   Free graded left module on 1 generator over The exterior algebra of rank 0 over Integer Ring,
             Module endomorphism of Free graded left module on 0 generators over The exterior algebra of rank 0 over Integer Ring]

        TESTS::

            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: M = FPModule(A2, [0,1], [[Sq(2), Sq(1)]])
            sage: res2 = M.resolution(2)
            sage: [type(f) for f in res2]
            [<class '...SteenrodFreeModuleHomspace...'>,
             <class '...SteenrodFreeModuleHomspace...'>,
             <class '...SteenrodFreeModuleHomspace...'>]
        """
        def _print_progress(i, k):
            if verbose:
                print('Computing f_%d (%d/%d)' % (i, i, k))

        if k < 0:
            raise ValueError('the length of the resolution must be nonnegative')

        ret_complex = []

        # Epsilon: F_0 -> M
        F_0 = self._free_module()
        epsilon = Hom(F_0, self)(self.generators())
        ret_complex.append(epsilon)

        if k == 0:
            return ret_complex

        # f_1: F_1 -> F_0
        _print_progress(1, k)
        F_1 = self._j.domain()
        pres = Hom(F_1, F_0)(tuple([F_0(x.dense_coefficient_list())
                                    for x in self._j.values()]))

        ret_complex.append(pres)

        # f_i: F_i -> F_i-1, for i > 1
        for i in range(2, k + 1):
            _print_progress(i, k)

            f = ret_complex[i - 1]
            ret_complex.append(f._resolve_kernel(top_dim=top_dim,
                                                 verbose=verbose))

        return ret_complex



"""
Generic matrices
"""

# ***************************************************************************
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.matrix.matrix1 cimport Matrix as Matrix1

cdef class Matrix(Matrix1):
    cdef _det_by_minors(self, Py_ssize_t level)
    cdef _pf_bfl(self)
    cdef bint _is_positive_definite_or_semidefinite(self, bint semi) except -1
    cdef tuple _block_ldlt(self, bint classical)
    cpdef _echelon(self, str algorithm)
    cpdef _echelon_in_place(self, str algorithm)
    cpdef matrix_window(self, Py_ssize_t row=*, Py_ssize_t col=*, Py_ssize_t nrows=*, Py_ssize_t ncols=*, bint check=*)
